import numpy as np

from hdmf.data_utils import DataChunkIterator, DataChunk
from hdmf.testing import TestCase


class DataChunkIteratorTests(TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_none_iter(self):
        """Test that DataChunkIterator __init__ sets defaults correctly and all chunks and recommended shapes are None.
        """
        dci = DataChunkIterator(dtype=np.dtype('int'))
        self.assertIsNone(dci.maxshape)
        self.assertEqual(dci.dtype, np.dtype('int'))
        self.assertEqual(dci.buffer_size, 1)
        self.assertEqual(dci.iter_axis, 0)
        count = 0
        for chunk in dci:
            pass
        self.assertEqual(count, 0)
        self.assertIsNone(dci.recommended_data_shape())
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_list_none(self):
        """Test that DataChunkIterator has no dtype or chunks when given a list of None.
        """
        a = [None, None, None]
        with self.assertRaisesWith(Exception, 'Data type could not be determined. Please specify dtype in '
                                              'DataChunkIterator init.'):
            DataChunkIterator(a)

    def test_list_none_dtype(self):
        """Test that DataChunkIterator has the passed-in dtype and no chunks when given a list of None.
        """
        a = [None, None, None]
        dci = DataChunkIterator(a, dtype=np.dtype('int'))
        self.assertTupleEqual(dci.maxshape, (3,))
        self.assertEqual(dci.dtype, np.dtype('int'))
        count = 0
        for chunk in dci:
            pass
        self.assertEqual(count, 0)
        self.assertTupleEqual(dci.recommended_data_shape(), (3,))
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_numpy_iter_unbuffered_first_axis(self):
        """Test DataChunkIterator with numpy data, no buffering, and iterating on the first dimension.
        """
        a = np.arange(30).reshape(5, 2, 3)
        dci = DataChunkIterator(data=a, buffer_size=1)
        count = 0
        for chunk in dci:
            self.assertTupleEqual(chunk.shape, (1, 2, 3))
            count += 1
        self.assertEqual(count, 5)
        self.assertTupleEqual(dci.recommended_data_shape(), a.shape)
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_numpy_iter_unbuffered_middle_axis(self):
        """Test DataChunkIterator with numpy data, no buffering, and iterating on a middle dimension.
        """
        a = np.arange(30).reshape(5, 2, 3)
        dci = DataChunkIterator(data=a, buffer_size=1, iter_axis=1)
        count = 0
        for chunk in dci:
            self.assertTupleEqual(chunk.shape, (5, 1, 3))
            count += 1
        self.assertEqual(count, 2)
        self.assertTupleEqual(dci.recommended_data_shape(), a.shape)
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_numpy_iter_unbuffered_last_axis(self):
        """Test DataChunkIterator with numpy data, no buffering, and iterating on the last dimension.
        """
        a = np.arange(30).reshape(5, 2, 3)
        dci = DataChunkIterator(data=a, buffer_size=1, iter_axis=2)
        count = 0
        for chunk in dci:
            self.assertTupleEqual(chunk.shape, (5, 2, 1))
            count += 1
        self.assertEqual(count, 3)
        self.assertTupleEqual(dci.recommended_data_shape(), a.shape)
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_numpy_iter_buffered_first_axis(self):
        """Test DataChunkIterator with numpy data, buffering, and iterating on the first dimension.
        """
        a = np.arange(30).reshape(5, 2, 3)
        dci = DataChunkIterator(data=a, buffer_size=2)
        count = 0
        for chunk in dci:
            if count < 2:
                self.assertTupleEqual(chunk.shape, (2, 2, 3))
            else:
                self.assertTupleEqual(chunk.shape, (1, 2, 3))
            count += 1
        self.assertEqual(count, 3)
        self.assertTupleEqual(dci.recommended_data_shape(), a.shape)
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_numpy_iter_buffered_middle_axis(self):
        """Test DataChunkIterator with numpy data, buffering, and iterating on a middle dimension.
        """
        a = np.arange(45).reshape(5, 3, 3)
        dci = DataChunkIterator(data=a, buffer_size=2, iter_axis=1)
        count = 0
        for chunk in dci:
            if count < 1:
                self.assertTupleEqual(chunk.shape, (5, 2, 3))
            else:
                self.assertTupleEqual(chunk.shape, (5, 1, 3))
            count += 1
        self.assertEqual(count, 2)
        self.assertTupleEqual(dci.recommended_data_shape(), a.shape)
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_numpy_iter_buffered_last_axis(self):
        """Test DataChunkIterator with numpy data, buffering, and iterating on the last dimension.
        """
        a = np.arange(30).reshape(5, 2, 3)
        dci = DataChunkIterator(data=a, buffer_size=2, iter_axis=2)
        count = 0
        for chunk in dci:
            if count < 1:
                self.assertTupleEqual(chunk.shape, (5, 2, 2))
            else:
                self.assertTupleEqual(chunk.shape, (5, 2, 1))
            count += 1
        self.assertEqual(count, 2)
        self.assertTupleEqual(dci.recommended_data_shape(), a.shape)
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_numpy_iter_unmatched_buffer_size(self):
        a = np.arange(10)
        dci = DataChunkIterator(data=a, buffer_size=3)
        self.assertTupleEqual(dci.maxshape, a.shape)
        self.assertEqual(dci.dtype, a.dtype)
        count = 0
        for chunk in dci:
            if count < 3:
                self.assertTupleEqual(chunk.data.shape, (3,))
            else:
                self.assertTupleEqual(chunk.data.shape, (1,))
            count += 1
        self.assertEqual(count, 4)
        self.assertTupleEqual(dci.recommended_data_shape(), a.shape)
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_standard_iterator_unbuffered(self):
        dci = DataChunkIterator(data=range(10), buffer_size=1)
        self.assertEqual(dci.dtype, np.dtype(int))
        self.assertTupleEqual(dci.maxshape, (10,))
        self.assertTupleEqual(dci.recommended_data_shape(), (10,))  # Test before and after iteration
        count = 0
        for chunk in dci:
            self.assertTupleEqual(chunk.data.shape, (1,))
            count += 1
        self.assertEqual(count, 10)
        self.assertTupleEqual(dci.recommended_data_shape(), (10,))  # Test before and after iteration
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_standard_iterator_unmatched_buffersized(self):
        dci = DataChunkIterator(data=range(10), buffer_size=3)
        self.assertEqual(dci.dtype, np.dtype(int))
        self.assertTupleEqual(dci.maxshape, (10,))
        self.assertIsNone(dci.recommended_chunk_shape())
        self.assertTupleEqual(dci.recommended_data_shape(), (10,))  # Test before and after iteration
        count = 0
        for chunk in dci:
            if count < 3:
                self.assertTupleEqual(chunk.data.shape, (3,))
            else:
                self.assertTupleEqual(chunk.data.shape, (1,))
            count += 1
        self.assertEqual(count, 4)
        self.assertTupleEqual(dci.recommended_data_shape(), (10,))  # Test before and after iteration

    def test_multidimensional_list_first_axis(self):
        """Test DataChunkIterator with multidimensional list data, no buffering, and iterating on the first dimension.
        """
        a = np.arange(30).reshape(5, 2, 3).tolist()
        dci = DataChunkIterator(a)
        self.assertTupleEqual(dci.maxshape, (5, 2, 3))
        self.assertEqual(dci.dtype, np.dtype(int))
        count = 0
        for chunk in dci:
            self.assertTupleEqual(chunk.data.shape, (1, 2, 3))
            count += 1
        self.assertEqual(count, 5)
        self.assertTupleEqual(dci.recommended_data_shape(), (5, 2, 3))
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_multidimensional_list_middle_axis(self):
        """Test DataChunkIterator with multidimensional list data, no buffering, and iterating on a middle dimension.
        """
        a = np.arange(30).reshape(5, 2, 3).tolist()
        warn_msg = ('Iterating over an axis other than the first dimension of list or tuple data '
                    'involves converting the data object to a numpy ndarray, which may incur a computational '
                    'cost.')
        with self.assertWarnsWith(UserWarning, warn_msg):
            dci = DataChunkIterator(a, iter_axis=1)
        self.assertTupleEqual(dci.maxshape, (5, 2, 3))
        self.assertEqual(dci.dtype, np.dtype(int))
        count = 0
        for chunk in dci:
            self.assertTupleEqual(chunk.data.shape, (5, 1, 3))
            count += 1
        self.assertEqual(count, 2)
        self.assertTupleEqual(dci.recommended_data_shape(), (5, 2, 3))
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_multidimensional_list_last_axis(self):
        """Test DataChunkIterator with multidimensional list data, no buffering, and iterating on the last dimension.
        """
        a = np.arange(30).reshape(5, 2, 3).tolist()
        warn_msg = ('Iterating over an axis other than the first dimension of list or tuple data '
                    'involves converting the data object to a numpy ndarray, which may incur a computational '
                    'cost.')
        with self.assertWarnsWith(UserWarning, warn_msg):
            dci = DataChunkIterator(a, iter_axis=2)
        self.assertTupleEqual(dci.maxshape, (5, 2, 3))
        self.assertEqual(dci.dtype, np.dtype(int))
        count = 0
        for chunk in dci:
            self.assertTupleEqual(chunk.data.shape, (5, 2, 1))
            count += 1
        self.assertEqual(count, 3)
        self.assertTupleEqual(dci.recommended_data_shape(), (5, 2, 3))
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_maxshape(self):
        a = np.arange(30).reshape(5, 2, 3)
        aiter = iter(a)
        daiter = DataChunkIterator.from_iterable(aiter, buffer_size=2)
        self.assertEqual(daiter.maxshape, (None, 2, 3))

    def test_dtype(self):
        a = np.arange(30, dtype='int32').reshape(5, 2, 3)
        aiter = iter(a)
        daiter = DataChunkIterator.from_iterable(aiter, buffer_size=2)
        self.assertEqual(daiter.dtype, a.dtype)

    def test_sparse_data_buffer_aligned(self):
        a = [1, 2, 3, 4, None, None, 7, 8, None, None]
        dci = DataChunkIterator(a, buffer_size=2)
        self.assertTupleEqual(dci.maxshape, (10,))
        self.assertEqual(dci.dtype, np.dtype(int))
        count = 0
        for chunk in dci:
            self.assertTupleEqual(chunk.data.shape, (2,))
            self.assertEqual(len(chunk.selection), 1)
            self.assertEqual(chunk.selection[0], slice(chunk.data[0] - 1, chunk.data[1]))
            count += 1
        self.assertEqual(count, 3)
        self.assertTupleEqual(dci.recommended_data_shape(), (10,))
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_sparse_data_buffer_notaligned(self):
        a = [1, 2, 3, None, None, None, None, 8, 9, 10]
        dci = DataChunkIterator(a, buffer_size=2)
        self.assertTupleEqual(dci.maxshape, (10,))
        self.assertEqual(dci.dtype, np.dtype(int))
        count = 0
        for chunk in dci:
            self.assertEqual(len(chunk.selection), 1)
            if count == 0:  # [1, 2]
                self.assertListEqual(chunk.data.tolist(), [1, 2])
                self.assertEqual(chunk.selection[0], slice(chunk.data[0] - 1, chunk.data[1]))
            elif count == 1:  # [3, None]
                self.assertListEqual(chunk.data.tolist(), [3, ])
                self.assertEqual(chunk.selection[0], slice(chunk.data[0] - 1, chunk.data[0]))
            elif count == 2:  # [8, 9]
                self.assertListEqual(chunk.data.tolist(), [8, 9])
                self.assertEqual(chunk.selection[0], slice(chunk.data[0] - 1, chunk.data[1]))
            else:  # count == 3, [10]
                self.assertListEqual(chunk.data.tolist(), [10, ])
                self.assertEqual(chunk.selection[0], slice(chunk.data[0] - 1, chunk.data[0]))
            count += 1
        self.assertEqual(count, 4)
        self.assertTupleEqual(dci.recommended_data_shape(), (10,))
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_start_with_none(self):
        a = [None, None, 3]
        dci = DataChunkIterator(a, buffer_size=2)
        self.assertTupleEqual(dci.maxshape, (3,))
        self.assertEqual(dci.dtype, np.dtype(int))
        count = 0
        for chunk in dci:
            self.assertListEqual(chunk.data.tolist(), [3])
            self.assertEqual(len(chunk.selection), 1)
            self.assertEqual(chunk.selection[0], slice(2, 3))
            count += 1
        self.assertEqual(count, 1)
        self.assertTupleEqual(dci.recommended_data_shape(), (3,))
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_list_scalar(self):
        a = [3]
        dci = DataChunkIterator(a, buffer_size=2)
        self.assertTupleEqual(dci.maxshape, (1,))
        self.assertEqual(dci.dtype, np.dtype(int))
        count = 0
        for chunk in dci:
            self.assertListEqual(chunk.data.tolist(), [3])
            self.assertEqual(len(chunk.selection), 1)
            self.assertEqual(chunk.selection[0], slice(0, 1))
            count += 1
        self.assertEqual(count, 1)
        self.assertTupleEqual(dci.recommended_data_shape(), (1,))
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_list_numpy_scalar(self):
        a = np.array([3])
        dci = DataChunkIterator(a, buffer_size=2)
        self.assertTupleEqual(dci.maxshape, (1,))
        self.assertEqual(dci.dtype, np.dtype(int))
        count = 0
        for chunk in dci:
            self.assertListEqual(chunk.data.tolist(), [3])
            self.assertEqual(len(chunk.selection), 1)
            self.assertEqual(chunk.selection[0], slice(0, 1))
            count += 1
        self.assertEqual(count, 1)
        self.assertTupleEqual(dci.recommended_data_shape(), (1,))
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_set_maxshape(self):
        a = np.array([3])
        dci = DataChunkIterator(a, maxshape=(5, 2, 3), buffer_size=2)
        self.assertTupleEqual(dci.maxshape, (5, 2, 3))
        self.assertEqual(dci.dtype, np.dtype(int))
        count = 0
        for chunk in dci:
            self.assertListEqual(chunk.data.tolist(), [3])
            self.assertTupleEqual(chunk.selection, (slice(0, 1), slice(None), slice(None)))
            count += 1
        self.assertEqual(count, 1)
        self.assertTupleEqual(dci.recommended_data_shape(), (5, 2, 3))
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_custom_iter_first_axis(self):
        def my_iter():
            count = 0
            a = np.arange(30).reshape(5, 2, 3)
            while count < a.shape[0]:
                val = a[count, :, :]
                count = count + 1
                yield val
            return
        dci = DataChunkIterator(data=my_iter(), buffer_size=2)
        count = 0
        for chunk in dci:
            if count < 2:
                self.assertTupleEqual(chunk.shape, (2, 2, 3))
            else:
                self.assertTupleEqual(chunk.shape, (1, 2, 3))
            count += 1
        self.assertEqual(count, 3)
        # self.assertTupleEqual(dci.recommended_data_shape(), (2, 2, 3))
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_custom_iter_middle_axis(self):
        def my_iter():
            count = 0
            a = np.arange(45).reshape(5, 3, 3)
            while count < a.shape[1]:
                val = a[:, count, :]
                count = count + 1
                yield val
            return
        dci = DataChunkIterator(data=my_iter(), buffer_size=2, iter_axis=1)
        count = 0
        for chunk in dci:
            if count < 1:
                self.assertTupleEqual(chunk.shape, (5, 2, 3))
            else:
                self.assertTupleEqual(chunk.shape, (5, 1, 3))
            count += 1
        self.assertEqual(count, 2)
        # self.assertTupleEqual(dci.recommended_data_shape(), (5, 2, 3))
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_custom_iter_last_axis(self):
        def my_iter():
            count = 0
            a = np.arange(30).reshape(5, 2, 3)
            while count < a.shape[2]:
                val = a[:, :, count]
                count = count + 1
                yield val
            return
        dci = DataChunkIterator(data=my_iter(), buffer_size=2, iter_axis=2)
        count = 0
        for chunk in dci:
            if count < 1:
                self.assertTupleEqual(chunk.shape, (5, 2, 2))
            else:
                self.assertTupleEqual(chunk.shape, (5, 2, 1))
            count += 1
        self.assertEqual(count, 2)
        # self.assertTupleEqual(dci.recommended_data_shape(), (5, 2, 2))
        self.assertIsNone(dci.recommended_chunk_shape())

    def test_custom_iter_mismatched_axis(self):
        def my_iter():
            count = 0
            a = np.arange(30).reshape(5, 2, 3)
            while count < a.shape[2]:
                val = a[:, :, count]
                count = count + 1
                yield val
            return
        # iterator returns slices of size (5, 2)
        # because iter_axis is by default 0, these chunks will be placed along the first dimension
        dci = DataChunkIterator(data=my_iter(), buffer_size=2)
        count = 0
        for chunk in dci:
            if count < 1:
                self.assertTupleEqual(chunk.shape, (2, 5, 2))
            else:
                self.assertTupleEqual(chunk.shape, (1, 5, 2))
            count += 1
        self.assertEqual(count, 2)
        # self.assertTupleEqual(dci.recommended_data_shape(), (5, 2, 2))
        self.assertIsNone(dci.recommended_chunk_shape())


class DataChunkTests(TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_len_operator_no_data(self):
        temp = DataChunk()
        self.assertEqual(len(temp), 0)

    def test_len_operator_with_data(self):
        temp = DataChunk(np.arange(10).reshape(5, 2))
        self.assertEqual(len(temp), 5)

    def test_dtype(self):
        temp = DataChunk(np.arange(10).astype('int'))
        temp_dtype = temp.dtype
        self.assertEqual(temp_dtype, np.dtype('int'))

    def test_astype(self):
        temp1 = DataChunk(np.arange(10).reshape(5, 2))
        temp2 = temp1.astype('float32')
        self.assertEqual(temp2.dtype, np.dtype('float32'))



import numpy as np
from copy import copy, deepcopy

from hdmf.data_utils import DataIO
from hdmf.container import Data
from hdmf.testing import TestCase


class DataIOTests(TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_copy(self):
        obj = DataIO(data=[1., 2., 3.])
        obj_copy = copy(obj)
        self.assertNotEqual(id(obj), id(obj_copy))
        self.assertEqual(id(obj.data), id(obj_copy.data))

    def test_deepcopy(self):
        obj = DataIO(data=[1., 2., 3.])
        obj_copy = deepcopy(obj)
        self.assertNotEqual(id(obj), id(obj_copy))
        self.assertNotEqual(id(obj.data), id(obj_copy.data))

    def test_dataio_slice_delegation(self):
        indata = np.arange(30)
        dset = DataIO(indata)
        self.assertTrue(np.all(dset[2:15] == indata[2:15]))

        indata = np.arange(50).reshape(5, 10)
        dset = DataIO(indata)
        self.assertTrue(np.all(dset[1:3, 5:8] == indata[1:3, 5:8]))

    def test_set_dataio(self):
        """
        Test that Data.set_dataio works as intended
        """
        dataio = DataIO()
        data = np.arange(30).reshape(5, 2, 3)
        container = Data('wrapped_data', data)
        container.set_dataio(dataio)
        self.assertIs(dataio.data, data)
        self.assertIs(dataio, container.data)

    def test_set_dataio_data_already_set(self):
        """
        Test that Data.set_dataio works as intended
        """
        dataio = DataIO(data=np.arange(30).reshape(5, 2, 3))
        data = np.arange(30).reshape(5, 2, 3)
        container = Data('wrapped_data', data)
        with self.assertRaisesWith(ValueError, "cannot overwrite 'data' on DataIO"):
            container.set_dataio(dataio)



from collections import OrderedDict
from datetime import datetime
from copy import copy
import ruamel.yaml as yaml
import os.path
import string
from warnings import warn
from itertools import chain
from abc import ABCMeta, abstractmethod

from ..utils import docval, getargs, popargs, get_docval, call_docval_func
from .catalog import SpecCatalog
from .spec import DatasetSpec, GroupSpec


_namespace_args = [
    {'name': 'doc', 'type': str, 'doc': 'a description about what this namespace represents'},
    {'name': 'name', 'type': str, 'doc': 'the name of this namespace'},
    {'name': 'schema', 'type': list, 'doc': 'location of schema specification files or other Namespaces'},
    {'name': 'full_name', 'type': str, 'doc': 'extended full name of this namespace', 'default': None},
    {'name': 'version', 'type': (str, tuple, list), 'doc': 'Version number of the namespace', 'default': None},
    {'name': 'date', 'type': (datetime, str),
     'doc': "Date last modified or released. Formatting is %Y-%m-%d %H:%M:%S, e.g, 2017-04-25 17:14:13",
     'default': None},
    {'name': 'author', 'type': (str, list), 'doc': 'Author or list of authors.', 'default': None},
    {'name': 'contact', 'type': (str, list),
     'doc': 'List of emails. Ordering should be the same as for author', 'default': None},
    {'name': 'catalog', 'type': SpecCatalog, 'doc': 'The SpecCatalog object for this SpecNamespace', 'default': None}
]


class SpecNamespace(dict):
    """
    A namespace for specifications
    """

    __types_key = 'data_types'

    UNVERSIONED = None  # value representing missing version

    @docval(*_namespace_args)
    def __init__(self, **kwargs):
        doc, full_name, name, version, date, author, contact, schema, catalog = \
            popargs('doc', 'full_name', 'name', 'version', 'date', 'author', 'contact', 'schema', 'catalog', kwargs)
        super().__init__()
        self['doc'] = doc
        self['schema'] = schema
        if any(c in string.whitespace for c in name):
            raise ValueError("'name' must not contain any whitespace")
        self['name'] = name
        if full_name is not None:
            self['full_name'] = full_name
        if version == str(SpecNamespace.UNVERSIONED):
            # the unversioned version may be written to file as a string and read from file as a string
            warn("Loaded namespace '%s' is unversioned. Please notify the extension author." % name)
            version = SpecNamespace.UNVERSIONED
        if version is None:
            # version is required on write -- see YAMLSpecWriter.write_namespace -- but can be None on read in order to
            # be able to read older files with extensions that are missing the version key.
            warn(("Loaded namespace '%s' is missing the required key 'version'. Version will be set to '%s'. "
                  "Please notify the extension author.") % (name, SpecNamespace.UNVERSIONED))
            version = SpecNamespace.UNVERSIONED
        self['version'] = version
        if date is not None:
            self['date'] = date
        if author is not None:
            self['author'] = author
        if contact is not None:
            self['contact'] = contact
        self.__catalog = catalog if catalog is not None else SpecCatalog()

    @classmethod
    def types_key(cls):
        ''' Get the key used for specifying types to include from a file or namespace

        Override this method to use a different name for 'data_types'
        '''
        return cls.__types_key

    @property
    def full_name(self):
        """String with full name or None"""
        return self.get('full_name', None)

    @property
    def contact(self):
        """String or list of strings with the contacts or None"""
        return self.get('contact', None)

    @property
    def author(self):
        """String or list of strings with the authors or None"""
        return self.get('author', None)

    @property
    def version(self):
        """
        String, list, or tuple with the version or SpecNamespace.UNVERSIONED
        if the version is missing or empty
        """
        return self.get('version', None) or SpecNamespace.UNVERSIONED

    @property
    def date(self):
        """Date last modified or released.

        :return: datetime object, string, or None"""
        return self.get('date', None)

    @property
    def name(self):
        """String with short name or None"""
        return self.get('name', None)

    @property
    def doc(self):
        return self['doc']

    @property
    def schema(self):
        return self['schema']

    def get_source_files(self):
        """
        Get the list of names of the source files included the schema of the namespace
        """
        return [item['source'] for item in self.schema if 'source' in item]

    @docval({'name': 'sourcefile', 'type': str, 'doc': 'Name of the source file'},
            returns='Dict with the source file documentation', rtype=dict)
    def get_source_description(self, sourcefile):
        """
        Get the description of a source file as described in the namespace. The result is a
        dict which contains the 'source' and optionally 'title', 'doc' and 'data_types'
        imported from the source file
        """
        for item in self.schema:
            if item.get('source', None) == sourcefile:
                return item

    @property
    def catalog(self):
        """The SpecCatalog containing all the Specs"""
        return self.__catalog

    @docval({'name': 'data_type', 'type': (str, type), 'doc': 'the data_type to get the spec for'})
    def get_spec(self, **kwargs):
        """Get the Spec object for the given data type"""
        data_type = getargs('data_type', kwargs)
        spec = self.__catalog.get_spec(data_type)
        if spec is None:
            raise ValueError("No specification for '%s' in namespace '%s'" % (data_type, self.name))
        return spec

    @docval(returns="the a tuple of the available data types", rtype=tuple)
    def get_registered_types(self, **kwargs):
        """Get the available types in this namespace"""
        return self.__catalog.get_registered_types()

    @docval({'name': 'data_type', 'type': (str, type), 'doc': 'the data_type to get the hierarchy of'},
            returns="a tuple with the type hierarchy", rtype=tuple)
    def get_hierarchy(self, **kwargs):
        ''' Get the extension hierarchy for the given data_type in this namespace'''
        data_type = getargs('data_type', kwargs)
        return self.__catalog.get_hierarchy(data_type)

    @classmethod
    def build_namespace(cls, **spec_dict):
        kwargs = copy(spec_dict)
        try:
            args = [kwargs.pop(x['name']) for x in get_docval(cls.__init__) if 'default' not in x]
        except KeyError as e:
            raise KeyError("'%s' not found in %s" % (e.args[0], str(spec_dict)))
        return cls(*args, **kwargs)


class SpecReader(metaclass=ABCMeta):

    @docval({'name': 'source', 'type': str, 'doc': 'the source from which this reader reads from'})
    def __init__(self, **kwargs):
        self.__source = getargs('source', kwargs)

    @property
    def source(self):
        return self.__source

    @abstractmethod
    def read_spec(self):
        pass

    @abstractmethod
    def read_namespace(self):
        pass


class YAMLSpecReader(SpecReader):

    @docval({'name': 'indir', 'type': str, 'doc': 'the path spec files are relative to', 'default': '.'})
    def __init__(self, **kwargs):
        super_kwargs = {'source': kwargs['indir']}
        call_docval_func(super().__init__, super_kwargs)

    def read_namespace(self, namespace_path):
        namespaces = None
        with open(namespace_path, 'r') as stream:
            d = yaml.safe_load(stream)
            namespaces = d.get('namespaces')
            if namespaces is None:
                raise ValueError("no 'namespaces' found in %s" % namespace_path)
        return namespaces

    def read_spec(self, spec_path):
        specs = None
        with open(self.__get_spec_path(spec_path), 'r') as stream:
            specs = yaml.safe_load(stream)
            if not ('datasets' in specs or 'groups' in specs):
                raise ValueError("no 'groups' or 'datasets' found in %s" % spec_path)
        return specs

    def __get_spec_path(self, spec_path):
        if os.path.isabs(spec_path):
            return spec_path
        return os.path.join(self.source, spec_path)


class NamespaceCatalog:

    @docval({'name': 'group_spec_cls', 'type': type,
             'doc': 'the class to use for group specifications', 'default': GroupSpec},
            {'name': 'dataset_spec_cls', 'type': type,
             'doc': 'the class to use for dataset specifications', 'default': DatasetSpec},
            {'name': 'spec_namespace_cls', 'type': type,
             'doc': 'the class to use for specification namespaces', 'default': SpecNamespace},)
    def __init__(self, **kwargs):
        """Create a catalog for storing  multiple Namespaces"""
        self.__namespaces = OrderedDict()
        self.__dataset_spec_cls = getargs('dataset_spec_cls', kwargs)
        self.__group_spec_cls = getargs('group_spec_cls', kwargs)
        self.__spec_namespace_cls = getargs('spec_namespace_cls', kwargs)
        # keep track of all spec objects ever loaded, so we don't have
        # multiple object instances of a spec
        self.__loaded_specs = dict()
        self.__included_specs = dict()
        self.__included_sources = dict()

        self._loaded_specs = self.__loaded_specs

    def __copy__(self):
        ret = NamespaceCatalog(self.__group_spec_cls,
                               self.__dataset_spec_cls,
                               self.__spec_namespace_cls)
        ret.__namespaces = copy(self.__namespaces)
        ret.__loaded_specs = copy(self.__loaded_specs)
        ret.__included_specs = copy(self.__included_specs)
        ret.__included_sources = copy(self.__included_sources)
        return ret

    def merge(self, ns_catalog):
        for name, namespace in ns_catalog.__namespaces.items():
            self.add_namespace(name, namespace)

    @property
    @docval(returns='a tuple of the available namespaces', rtype=tuple)
    def namespaces(self):
        """The namespaces in this NamespaceCatalog"""
        return tuple(self.__namespaces.keys())

    @property
    def dataset_spec_cls(self):
        """The DatasetSpec class used in this NamespaceCatalog"""
        return self.__dataset_spec_cls

    @property
    def group_spec_cls(self):
        """The GroupSpec class used in this NamespaceCatalog"""
        return self.__group_spec_cls

    @property
    def spec_namespace_cls(self):
        """The SpecNamespace class used in this NamespaceCatalog"""
        return self.__spec_namespace_cls

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this namespace'},
            {'name': 'namespace', 'type': SpecNamespace, 'doc': 'the SpecNamespace object'})
    def add_namespace(self, **kwargs):
        """Add a namespace to this catalog"""
        name, namespace = getargs('name', 'namespace', kwargs)
        if name in self.__namespaces:
            raise KeyError("namespace '%s' already exists" % name)
        self.__namespaces[name] = namespace
        for dt in namespace.catalog.get_registered_types():
            source = namespace.catalog.get_spec_source_file(dt)
            self.__loaded_specs.setdefault(source, list()).append(dt)

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this namespace'},
            returns="the SpecNamespace with the given name", rtype=SpecNamespace)
    def get_namespace(self, **kwargs):
        """Get the a SpecNamespace"""
        name = getargs('name', kwargs)
        ret = self.__namespaces.get(name)
        if ret is None:
            raise KeyError("'%s' not a namespace" % name)
        return ret

    @docval({'name': 'namespace', 'type': str, 'doc': 'the name of the namespace'},
            {'name': 'data_type', 'type': (str, type), 'doc': 'the data_type to get the spec for'},
            returns="the specification for writing the given object type to HDF5 ", rtype='Spec')
    def get_spec(self, **kwargs):
        '''
        Get the Spec object for the given type from the given Namespace
        '''
        namespace, data_type = getargs('namespace', 'data_type', kwargs)
        if namespace not in self.__namespaces:
            raise KeyError("'%s' not a namespace" % namespace)
        return self.__namespaces[namespace].get_spec(data_type)

    @docval({'name': 'namespace', 'type': str, 'doc': 'the name of the namespace'},
            {'name': 'data_type', 'type': (str, type), 'doc': 'the data_type to get the spec for'},
            returns="a tuple with the type hierarchy", rtype=tuple)
    def get_hierarchy(self, **kwargs):
        '''
        Get the type hierarchy for a given data_type in a given namespace
        '''
        namespace, data_type = getargs('namespace', 'data_type', kwargs)
        spec_ns = self.__namespaces.get(namespace)
        if spec_ns is None:
            raise KeyError("'%s' not a namespace" % namespace)
        return spec_ns.get_hierarchy(data_type)

    @docval(rtype=tuple)
    def get_sources(self, **kwargs):
        '''
        Get all the source specification files that were loaded in this catalog
        '''
        return tuple(self.__loaded_specs.keys())

    @docval({'name': 'namespace', 'type': str, 'doc': 'the name of the namespace'},
            rtype=tuple)
    def get_namespace_sources(self, **kwargs):
        '''
        Get all the source specifications that were loaded for a given namespace
        '''
        namespace = getargs('namespace', kwargs)
        return tuple(self.__included_sources[namespace])

    @docval({'name': 'source', 'type': str, 'doc': 'the name of the source'},
            rtype=tuple)
    def get_types(self, **kwargs):
        '''
        Get the types that were loaded from a given source
        '''
        source = getargs('source', kwargs)
        ret = self.__loaded_specs.get(source)
        if ret is not None:
            ret = tuple(ret)
        return ret

    def __load_spec_file(self, reader, spec_source, catalog, dtypes=None, resolve=True):
        ret = self.__loaded_specs.get(spec_source)
        if ret is not None:
            raise ValueError("spec source '%s' already loaded" % spec_source)

        def __reg_spec(spec_cls, spec_dict):
            parent_cls = GroupSpec if issubclass(spec_cls, GroupSpec) else DatasetSpec
            dt_def = spec_dict.get(spec_cls.def_key(), spec_dict.get(parent_cls.def_key()))
            if dt_def is None:
                msg = 'no %s or %s found in spec %s' % (spec_cls.def_key(), parent_cls.def_key(), spec_source)
                raise ValueError(msg)
            if dtypes and dt_def not in dtypes:
                return
            if resolve:
                self.__resolve_includes(spec_dict, catalog)
            spec_obj = spec_cls.build_spec(spec_dict)
            return catalog.auto_register(spec_obj, spec_source)

        if ret is None:
            ret = list()
            d = reader.read_spec(spec_source)
            specs = d.get('datasets', list())
            for spec_dict in specs:
                ret.extend(__reg_spec(self.__dataset_spec_cls, spec_dict))
            specs = d.get('groups', list())
            for spec_dict in specs:
                ret.extend(__reg_spec(self.__group_spec_cls, spec_dict))
            self.__loaded_specs[spec_source] = ret
        return ret

    def __resolve_includes(self, spec_dict, catalog):
        """
            Pull in any attributes, datasets, or groups included
        """
        dt_inc = spec_dict.get(self.__group_spec_cls.inc_key())
        dt_def = spec_dict.get(self.__group_spec_cls.def_key())
        if dt_inc is not None and dt_def is not None:
            parent_spec = catalog.get_spec(dt_inc)
            if parent_spec is None:
                msg = "Cannot resolve include spec '%s' for type '%s'" % (dt_inc, dt_def)
                raise ValueError(msg)
            spec_dict[self.__group_spec_cls.inc_key()] = parent_spec
        it = chain(spec_dict.get('groups', list()), spec_dict.get('datasets', list()))
        for subspec_dict in it:
            self.__resolve_includes(subspec_dict, catalog)

    def __load_namespace(self, namespace, reader, types_key, resolve=True):
        ns_name = namespace['name']
        if ns_name in self.__namespaces:
            raise KeyError("namespace '%s' already exists" % ns_name)
        catalog = SpecCatalog()
        included_types = dict()
        for s in namespace['schema']:
            if 'source' in s:
                # read specs from file
                dtypes = None
                if types_key in s:
                    dtypes = set(s[types_key])
                self.__load_spec_file(reader, s['source'], catalog, dtypes=dtypes, resolve=resolve)
                self.__included_sources.setdefault(ns_name, list()).append(s['source'])
            elif 'namespace' in s:
                # load specs from namespace
                try:
                    inc_ns = self.get_namespace(s['namespace'])
                except KeyError as e:
                    raise ValueError("Could not load namespace '%s'" % s['namespace']) from e
                if types_key in s:
                    types = s[types_key]
                else:
                    types = inc_ns.get_registered_types()
                for ndt in types:
                    spec = inc_ns.get_spec(ndt)
                    spec_file = inc_ns.catalog.get_spec_source_file(ndt)
                    if isinstance(spec, DatasetSpec):
                        spec = self.dataset_spec_cls.build_spec(spec)
                    else:
                        spec = self.group_spec_cls.build_spec(spec)
                    catalog.register_spec(spec, spec_file)
                included_types[s['namespace']] = tuple(types)
        # construct namespace
        ns = self.__spec_namespace_cls.build_namespace(catalog=catalog, **namespace)
        self.__namespaces[ns_name] = ns
        return included_types

    @docval({'name': 'namespace_path', 'type': str, 'doc': 'the path to the file containing the namespaces(s) to load'},
            {'name': 'resolve',
             'type': bool,
             'doc': 'whether or not to include objects from included/parent spec objects', 'default': True},
            {'name': 'reader',
             'type': SpecReader,
             'doc': 'the class to user for reading specifications', 'default': None},
            returns='a dictionary describing the dependencies of loaded namespaces', rtype=dict)
    def load_namespaces(self, **kwargs):
        """Load the namespaces in the given file"""
        namespace_path, resolve, reader = getargs('namespace_path', 'resolve', 'reader', kwargs)
        if reader is None:
            # load namespace definition from file
            if not os.path.exists(namespace_path):
                msg = "namespace file '%s' not found" % namespace_path
                raise IOError(msg)
            reader = YAMLSpecReader(indir=os.path.dirname(namespace_path))
        ns_path_key = os.path.join(reader.source, os.path.basename(namespace_path))
        ret = self.__included_specs.get(ns_path_key)
        if ret is None:
            ret = dict()
        else:
            return ret
        namespaces = reader.read_namespace(namespace_path)
        types_key = self.__spec_namespace_cls.types_key()
        to_load = list()
        for ns in namespaces:
            if ns['name'] in self.__namespaces:
                warn("ignoring namespace '%s' because it already exists" % ns['name'])
            else:
                to_load.append(ns)
        # now load specs into namespace
        for ns in to_load:
            ret[ns['name']] = self.__load_namespace(ns, reader, types_key, resolve=resolve)
        self.__included_specs[ns_path_key] = ret
        return ret



import copy

from hdmf.spec import GroupSpec, DatasetSpec, AttributeSpec, SpecCatalog
from hdmf.testing import TestCase


class SpecCatalogTest(TestCase):

    def setUp(self):
        self.catalog = SpecCatalog()
        self.attributes = [
            AttributeSpec('attribute1', 'my first attribute', 'text'),
            AttributeSpec('attribute2', 'my second attribute', 'text')
        ]
        self.spec = DatasetSpec('my first dataset',
                                'int',
                                name='dataset1',
                                dims=(None, None),
                                attributes=self.attributes,
                                linkable=False,
                                data_type_def='EphysData')

    def test_register_spec(self):
        self.catalog.register_spec(self.spec, 'test.yaml')
        result = self.catalog.get_spec('EphysData')
        self.assertIs(result, self.spec)

    def test_hierarchy(self):
        spikes_spec = DatasetSpec('my extending dataset', 'int',
                                  data_type_inc='EphysData',
                                  data_type_def='SpikeData')

        lfp_spec = DatasetSpec('my second extending dataset', 'int',
                               data_type_inc='EphysData',
                               data_type_def='LFPData')

        self.catalog.register_spec(self.spec, 'test.yaml')
        self.catalog.register_spec(spikes_spec, 'test.yaml')
        self.catalog.register_spec(lfp_spec, 'test.yaml')

        spike_hierarchy = self.catalog.get_hierarchy('SpikeData')
        lfp_hierarchy = self.catalog.get_hierarchy('LFPData')
        ephys_hierarchy = self.catalog.get_hierarchy('EphysData')
        self.assertTupleEqual(spike_hierarchy, ('SpikeData', 'EphysData'))
        self.assertTupleEqual(lfp_hierarchy, ('LFPData', 'EphysData'))
        self.assertTupleEqual(ephys_hierarchy, ('EphysData',))

    def test_subtypes(self):
        """
         -BaseContainer--+-->AContainer--->ADContainer
                        |
                        +-->BContainer
        """
        base_spec = GroupSpec(doc='Base container',
                              data_type_def='BaseContainer')
        acontainer = GroupSpec(doc='AContainer',
                               data_type_inc='BaseContainer',
                               data_type_def='AContainer')
        adcontainer = GroupSpec(doc='ADContainer',
                                data_type_inc='AContainer',
                                data_type_def='ADContainer')
        bcontainer = GroupSpec(doc='BContainer',
                               data_type_inc='BaseContainer',
                               data_type_def='BContainer')
        self.catalog.register_spec(base_spec, 'test.yaml')
        self.catalog.register_spec(acontainer, 'test.yaml')
        self.catalog.register_spec(adcontainer, 'test.yaml')
        self.catalog.register_spec(bcontainer, 'test.yaml')
        base_spec_subtypes = self.catalog.get_subtypes('BaseContainer')
        base_spec_subtypes = tuple(sorted(base_spec_subtypes))  # Sort so we have a guaranteed order for comparison
        acontainer_subtypes = self.catalog.get_subtypes('AContainer')
        bcontainer_substypes = self.catalog.get_subtypes('BContainer')
        adcontainer_subtypes = self.catalog.get_subtypes('ADContainer')
        self.assertTupleEqual(adcontainer_subtypes, ())
        self.assertTupleEqual(bcontainer_substypes, ())
        self.assertTupleEqual(acontainer_subtypes, ('ADContainer',))
        self.assertTupleEqual(base_spec_subtypes,  ('AContainer', 'ADContainer', 'BContainer'))

    def test_subtypes_norecursion(self):
        """
         -BaseContainer--+-->AContainer--->ADContainer
                        |
                        +-->BContainer
        """
        base_spec = GroupSpec(doc='Base container',
                              data_type_def='BaseContainer')
        acontainer = GroupSpec(doc='AContainer',
                               data_type_inc='BaseContainer',
                               data_type_def='AContainer')
        adcontainer = GroupSpec(doc='ADContainer',
                                data_type_inc='AContainer',
                                data_type_def='ADContainer')
        bcontainer = GroupSpec(doc='BContainer',
                               data_type_inc='BaseContainer',
                               data_type_def='BContainer')
        self.catalog.register_spec(base_spec, 'test.yaml')
        self.catalog.register_spec(acontainer, 'test.yaml')
        self.catalog.register_spec(adcontainer, 'test.yaml')
        self.catalog.register_spec(bcontainer, 'test.yaml')
        base_spec_subtypes = self.catalog.get_subtypes('BaseContainer', recursive=False)
        base_spec_subtypes = tuple(sorted(base_spec_subtypes))  # Sort so we have a guaranteed order for comparison
        acontainer_subtypes = self.catalog.get_subtypes('AContainer', recursive=False)
        bcontainer_substypes = self.catalog.get_subtypes('BContainer', recursive=False)
        adcontainer_subtypes = self.catalog.get_subtypes('ADContainer', recursive=False)
        self.assertTupleEqual(adcontainer_subtypes, ())
        self.assertTupleEqual(bcontainer_substypes, ())
        self.assertTupleEqual(acontainer_subtypes, ('ADContainer',))
        self.assertTupleEqual(base_spec_subtypes,  ('AContainer', 'BContainer'))

    def test_subtypes_unknown_type(self):
        subtypes_of_bad_type = self.catalog.get_subtypes('UnknownType')
        self.assertTupleEqual(subtypes_of_bad_type, ())

    def test_get_spec_source_file(self):
        spikes_spec = GroupSpec('test group',
                                data_type_def='SpikeData')
        source_file_path = '/test/myt/test.yaml'
        self.catalog.auto_register(spikes_spec, source_file_path)
        recorded_source_file_path = self.catalog.get_spec_source_file('SpikeData')
        self.assertEqual(recorded_source_file_path, source_file_path)

    def test_get_full_hierarchy(self):
        """
        BaseContainer--+-->AContainer--->ADContainer
                        |
                        +-->BContainer

        Expected output:
        >> print(json.dumps(full_hierarchy, indent=4))
        >> {
        >>     "BaseContainer": {
        >>         "AContainer": {
        >>             "ADContainer": {}
        >>         },
        >>          "BContainer": {}
        >> }
        """
        base_spec = GroupSpec(doc='Base container',
                              data_type_def='BaseContainer')
        acontainer = GroupSpec(doc='AContainer',
                               data_type_inc='BaseContainer',
                               data_type_def='AContainer')
        adcontainer = GroupSpec(doc='ADContainer',
                                data_type_inc='AContainer',
                                data_type_def='ADContainer')
        bcontainer = GroupSpec(doc='BContainer',
                               data_type_inc='BaseContainer',
                               data_type_def='BContainer')
        self.catalog.register_spec(base_spec, 'test.yaml')
        self.catalog.register_spec(acontainer, 'test.yaml')
        self.catalog.register_spec(adcontainer, 'test.yaml')
        self.catalog.register_spec(bcontainer, 'test.yaml')
        full_hierarchy = self.catalog.get_full_hierarchy()
        expected_hierarchy = {
                                "BaseContainer": {
                                    "AContainer": {
                                        "ADContainer": {}
                                    },
                                    "BContainer": {}
                                }
                             }
        self.assertDictEqual(full_hierarchy, expected_hierarchy)

    def test_copy_spec_catalog(self):
        # Register the spec first
        self.catalog.register_spec(self.spec, 'test.yaml')
        result = self.catalog.get_spec('EphysData')
        self.assertIs(result, self.spec)
        # Now test the copy
        re = copy.copy(self.catalog)
        self.assertTupleEqual(self.catalog.get_registered_types(),
                              re.get_registered_types())

    def test_deepcopy_spec_catalog(self):
        # Register the spec first
        self.catalog.register_spec(self.spec, 'test.yaml')
        result = self.catalog.get_spec('EphysData')
        self.assertIs(result, self.spec)
        # Now test the copy
        re = copy.deepcopy(self.catalog)
        self.assertTupleEqual(self.catalog.get_registered_types(),
                              re.get_registered_types())



import numpy as np

from hdmf.data_utils import ShapeValidatorResult, DataChunkIterator, assertEqualShape
from hdmf.common.table import DynamicTable, DynamicTableRegion, VectorData
from hdmf.testing import TestCase


class ShapeValidatorTests(TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_array_all_dimensions_match(self):
        # Test match
        d1 = np.arange(10).reshape(2, 5)
        d2 = np.arange(10).reshape(2, 5)
        res = assertEqualShape(d1, d2)
        self.assertTrue(res.result)
        self.assertIsNone(res.error)
        self.assertTupleEqual(res.ignored, ())
        self.assertTupleEqual(res.unmatched, ())
        self.assertTupleEqual(res.shape1, (2, 5))
        self.assertTupleEqual(res.shape2, (2, 5))
        self.assertTupleEqual(res.axes1, (0, 1))
        self.assertTupleEqual(res.axes2, (0, 1))

    def test_array_dimensions_mismatch(self):
        # Test unmatched
        d1 = np.arange(10).reshape(2, 5)
        d2 = np.arange(10).reshape(5, 2)
        res = assertEqualShape(d1, d2)
        self.assertFalse(res.result)
        self.assertEqual(res.error, 'AXIS_LEN_ERROR')
        self.assertTupleEqual(res.ignored, ())
        self.assertTupleEqual(res.unmatched, ((0, 0), (1, 1)))
        self.assertTupleEqual(res.shape1, (2, 5))
        self.assertTupleEqual(res.shape2, (5, 2))
        self.assertTupleEqual(res.axes1, (0, 1))
        self.assertTupleEqual(res.axes2, (0, 1))

    def test_array_unequal_number_of_dimensions(self):
        # Test unequal num dims
        d1 = np.arange(10).reshape(2, 5)
        d2 = np.arange(20).reshape(5, 2, 2)
        res = assertEqualShape(d1, d2)
        self.assertFalse(res.result)
        self.assertEqual(res.error, 'NUM_AXES_ERROR')
        self.assertTupleEqual(res.ignored, ())
        self.assertTupleEqual(res.unmatched, ())
        self.assertTupleEqual(res.shape1, (2, 5))
        self.assertTupleEqual(res.shape2, (5, 2, 2))
        self.assertTupleEqual(res.axes1, (0, 1))
        self.assertTupleEqual(res.axes2, (0, 1, 2))

    def test_array_unequal_number_of_dimensions_check_one_axis_only(self):
        # Test unequal num dims compare one axis
        d1 = np.arange(10).reshape(2, 5)
        d2 = np.arange(20).reshape(2, 5, 2)
        res = assertEqualShape(d1, d2, 0, 0)
        self.assertTrue(res.result)
        self.assertIsNone(res.error)
        self.assertTupleEqual(res.ignored, ())
        self.assertTupleEqual(res.unmatched, ())
        self.assertTupleEqual(res.shape1, (2, 5))
        self.assertTupleEqual(res.shape2, (2, 5, 2))
        self.assertTupleEqual(res.axes1, (0,))
        self.assertTupleEqual(res.axes2, (0,))

    def test_array_unequal_number_of_dimensions_check_multiple_axesy(self):
        # Test unequal num dims compare multiple axes
        d1 = np.arange(10).reshape(2, 5)
        d2 = np.arange(20).reshape(5, 2, 2)
        res = assertEqualShape(d1, d2, [0, 1], [1, 0])
        self.assertTrue(res.result)
        self.assertIsNone(res.error)
        self.assertTupleEqual(res.ignored, ())
        self.assertTupleEqual(res.unmatched, ())
        self.assertTupleEqual(res.shape1, (2, 5))
        self.assertTupleEqual(res.shape2, (5, 2, 2))
        self.assertTupleEqual(res.axes1, (0, 1))
        self.assertTupleEqual(res.axes2, (1, 0))

    def test_array_unequal_number_of_axes_for_comparison(self):
        # Test unequal num axes for comparison
        d1 = np.arange(10).reshape(2, 5)
        d2 = np.arange(20).reshape(5, 2, 2)
        res = assertEqualShape(d1, d2, [0, 1], 1)
        self.assertFalse(res.result)
        self.assertEqual(res.error, "NUM_AXES_ERROR")
        self.assertTupleEqual(res.ignored, ())
        self.assertTupleEqual(res.unmatched, ())
        self.assertTupleEqual(res.shape1, (2, 5))
        self.assertTupleEqual(res.shape2, (5, 2, 2))
        self.assertTupleEqual(res.axes1, (0, 1))
        self.assertTupleEqual(res.axes2, (1,))

    def test_array_axis_index_out_of_bounds_single_axis(self):
        # Test too large frist axis
        d1 = np.arange(10).reshape(2, 5)
        d2 = np.arange(20).reshape(5, 2, 2)
        res = assertEqualShape(d1, d2, 4, 1)
        self.assertFalse(res.result)
        self.assertEqual(res.error, 'AXIS_OUT_OF_BOUNDS')
        self.assertTupleEqual(res.ignored, ())
        self.assertTupleEqual(res.unmatched, ())
        self.assertTupleEqual(res.shape1, (2, 5))
        self.assertTupleEqual(res.shape2, (5, 2, 2))
        self.assertTupleEqual(res.axes1, (4,))
        self.assertTupleEqual(res.axes2, (1,))

    def test_array_axis_index_out_of_bounds_mutilple_axis(self):
        # Test too large second axis
        d1 = np.arange(10).reshape(2, 5)
        d2 = np.arange(20).reshape(5, 2, 2)
        res = assertEqualShape(d1, d2, [0, 1], [5, 0])
        self.assertFalse(res.result)
        self.assertEqual(res.error, 'AXIS_OUT_OF_BOUNDS')
        self.assertTupleEqual(res.ignored, ())
        self.assertTupleEqual(res.unmatched, ())
        self.assertTupleEqual(res.shape1, (2, 5))
        self.assertTupleEqual(res.shape2, (5, 2, 2))
        self.assertTupleEqual(res.axes1, (0, 1))
        self.assertTupleEqual(res.axes2, (5, 0))

    def test_DataChunkIterators_match(self):
        # Compare data chunk iterators
        d1 = DataChunkIterator(data=np.arange(10).reshape(2, 5))
        d2 = DataChunkIterator(data=np.arange(10).reshape(2, 5))
        res = assertEqualShape(d1, d2)
        self.assertTrue(res.result)
        self.assertIsNone(res.error)
        self.assertTupleEqual(res.ignored, ())
        self.assertTupleEqual(res.unmatched, ())
        self.assertTupleEqual(res.shape1, (2, 5))
        self.assertTupleEqual(res.shape2, (2, 5))
        self.assertTupleEqual(res.axes1, (0, 1))
        self.assertTupleEqual(res.axes2, (0, 1))

    def test_DataChunkIterator_ignore_undetermined_axis(self):
        # Compare data chunk iterators with undetermined axis (ignore axis)
        d1 = DataChunkIterator(data=np.arange(10).reshape(2, 5), maxshape=(None, 5))
        d2 = DataChunkIterator(data=np.arange(10).reshape(2, 5))
        res = assertEqualShape(d1, d2, ignore_undetermined=True)
        self.assertTrue(res.result)
        self.assertIsNone(res.error)
        self.assertTupleEqual(res.ignored, ((0, 0),))
        self.assertTupleEqual(res.unmatched, ())
        self.assertTupleEqual(res.shape1, (None, 5))
        self.assertTupleEqual(res.shape2, (2, 5))
        self.assertTupleEqual(res.axes1, (0, 1))
        self.assertTupleEqual(res.axes2, (0, 1))

    def test_DataChunkIterator_error_on_undetermined_axis(self):
        # Compare data chunk iterators with undetermined axis (error on undetermined axis)
        d1 = DataChunkIterator(data=np.arange(10).reshape(2, 5), maxshape=(None, 5))
        d2 = DataChunkIterator(data=np.arange(10).reshape(2, 5))
        res = assertEqualShape(d1, d2, ignore_undetermined=False)
        self.assertFalse(res.result)
        self.assertEqual(res.error, 'AXIS_LEN_ERROR')
        self.assertTupleEqual(res.ignored, ())
        self.assertTupleEqual(res.unmatched, ((0, 0),))
        self.assertTupleEqual(res.shape1, (None, 5))
        self.assertTupleEqual(res.shape2, (2, 5))
        self.assertTupleEqual(res.axes1, (0, 1))
        self.assertTupleEqual(res.axes2, (0, 1))

    def test_DynamicTableRegion_shape_validation(self):
        # Create a test DynamicTable
        dt_spec = [
            {'name': 'foo', 'description': 'foo column'},
            {'name': 'bar', 'description': 'bar column'},
            {'name': 'baz', 'description': 'baz column'},
        ]
        dt_data = [
            [1, 2, 3, 4, 5],
            [10.0, 20.0, 30.0, 40.0, 50.0],
            ['cat', 'dog', 'bird', 'fish', 'lizard']
        ]
        columns = [
            VectorData(name=s['name'], description=s['description'], data=d)
            for s, d in zip(dt_spec, dt_data)
        ]
        dt = DynamicTable("with_columns_and_data",
                          "a test table", columns=columns)
        # Create test DynamicTableRegion
        dtr = DynamicTableRegion('dtr', [1, 2, 2], 'desc', table=dt)
        # Confirm that the shapes match
        res = assertEqualShape(dtr, np.arange(9).reshape(3, 3))
        self.assertTrue(res.result)

    def with_table_columns(self):
        cols = [VectorData(**d) for d in self.spec]
        table = DynamicTable("with_table_columns", 'a test table', columns=cols)
        return table

    def with_columns_and_data(self):

        return


class ShapeValidatorResultTests(TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_default_message(self):
        temp = ShapeValidatorResult()
        temp.error = 'AXIS_LEN_ERROR'
        self.assertEqual(temp.default_message, ShapeValidatorResult.SHAPE_ERROR[temp.error])

    def test_set_error_to_illegal_type(self):
        temp = ShapeValidatorResult()
        with self.assertRaises(ValueError):
            temp.error = 'MY_ILLEGAL_ERROR_TYPE'

    def test_ensure_use_of_tuples_during_asignment(self):
        temp = ShapeValidatorResult()
        temp_d = [1, 2]
        temp_cases = ['shape1', 'shape2', 'axes1', 'axes2', 'ignored', 'unmatched']
        for var in temp_cases:
            setattr(temp, var, temp_d)
            self.assertIsInstance(getattr(temp, var), tuple,  var)



from hdmf.common import CSRMatrix
from hdmf.testing import TestCase, H5RoundTripMixin

import scipy.sparse as sps
import numpy as np


class TestCSRMatrix(TestCase):

    def test_from_sparse_matrix(self):
        data = np.array([1, 2, 3, 4, 5, 6])
        indices = np.array([0, 2, 2, 0, 1, 2])
        indptr = np.array([0, 2, 3, 6])
        expected = CSRMatrix(data, indices, indptr, (3, 3))

        sps_mat = sps.csr_matrix((data, indices, indptr), shape=(3, 3))
        received = CSRMatrix(sps_mat)
        self.assertContainerEqual(received, expected, ignore_hdmf_attrs=True)

    def test_to_spmat(self):
        data = np.array([1, 2, 3, 4, 5, 6])
        indices = np.array([0, 2, 2, 0, 1, 2])
        indptr = np.array([0, 2, 3, 6])
        csr_mat = CSRMatrix(data, indices, indptr, (3, 3))
        spmat_array = csr_mat.to_spmat().toarray()

        expected = np.asarray([[1, 0, 2], [0, 0, 3], [4, 5, 6]])
        np.testing.assert_array_equal(spmat_array, expected)

    # TODO more unit tests are needed for CSRMatrix


class TestCSRMatrixRoundTrip(H5RoundTripMixin, TestCase):

    def setUpContainer(self):
        data = np.array([1, 2, 3, 4, 5, 6])
        indices = np.array([0, 2, 2, 0, 1, 2])
        indptr = np.array([0, 2, 3, 6])
        return CSRMatrix(data, indices, indptr, (3, 3))



'''This package will contain functions, classes, and objects
for reading and writing data in according to the HDMF-common specification
'''
import os.path
from copy import deepcopy

CORE_NAMESPACE = 'hdmf-common'

from ..spec import NamespaceCatalog  # noqa: E402
from ..utils import docval, getargs, call_docval_func  # noqa: E402
from ..backends.io import HDMFIO  # noqa: E402
from ..validate import ValidatorMap  # noqa: E402
from ..build import BuildManager, TypeMap  # noqa: E402


# a global type map
global __TYPE_MAP


# a function to register a container classes with the global map
@docval({'name': 'data_type', 'type': str, 'doc': 'the data_type to get the spec for'},
        {'name': 'namespace', 'type': str, 'doc': 'the name of the namespace', 'default': CORE_NAMESPACE},
        {"name": "container_cls", "type": type,
         "doc": "the class to map to the specified data_type", 'default': None},
        is_method=False)
def register_class(**kwargs):
    """Register an Container class to use for reading and writing a data_type from a specification
    If container_cls is not specified, returns a decorator for registering an Container subclass
    as the class for data_type in namespace.
    """
    data_type, namespace, container_cls = getargs('data_type', 'namespace', 'container_cls', kwargs)

    def _dec(cls):
        __TYPE_MAP.register_container_type(namespace, data_type, cls)
        return cls
    if container_cls is None:
        return _dec
    else:
        _dec(container_cls)


# a function to register an object mapper for a container class
@docval({"name": "container_cls", "type": type,
         "doc": "the Container class for which the given ObjectMapper class gets used for"},
        {"name": "mapper_cls", "type": type, "doc": "the ObjectMapper class to use to map", 'default': None},
        is_method=False)
def register_map(**kwargs):
    """Register an ObjectMapper to use for a Container class type
    If mapper_cls is not specified, returns a decorator for registering an ObjectMapper class
    as the mapper for container_cls. If mapper_cls specified, register the class as the mapper for container_cls
    """
    container_cls, mapper_cls = getargs('container_cls', 'mapper_cls', kwargs)

    def _dec(cls):
        __TYPE_MAP.register_map(container_cls, cls)
        return cls
    if mapper_cls is None:
        return _dec
    else:
        _dec(mapper_cls)


def __get_resources():
    from pkg_resources import resource_filename
    from os.path import join
    __core_ns_file_name = 'namespace.yaml'

    ret = dict()
    ret['namespace_path'] = join(resource_filename(__name__, 'hdmf-common-schema/common'), __core_ns_file_name)
    return ret


def _get_resources():
    # LEGACY: Needed to support legacy implementation.
    return __get_resources()


@docval({'name': 'namespace_path', 'type': str,
         'doc': 'the path to the YAML with the namespace definition'},
        returns="the namespaces loaded from the given file", rtype=tuple,
        is_method=False)
def load_namespaces(**kwargs):
    '''
    Load namespaces from file
    '''
    namespace_path = getargs('namespace_path', kwargs)
    return __TYPE_MAP.load_namespaces(namespace_path)


def available_namespaces():
    return __TYPE_MAP.namespace_catalog.namespaces


# load the hdmf-common namespace
__resources = __get_resources()
if os.path.exists(__resources['namespace_path']):
    __TYPE_MAP = TypeMap(NamespaceCatalog())

    load_namespaces(__resources['namespace_path'])

    # import these so the TypeMap gets populated
    from . import io as __io  # noqa: F401,E402

    from . import table  # noqa: F401,E402
    from . import sparse  # noqa: F401,E402

    from .. import Data, Container
    __TYPE_MAP.register_container_type(CORE_NAMESPACE, 'Container', Container)
    __TYPE_MAP.register_container_type(CORE_NAMESPACE, 'Data', Data)

else:
    raise RuntimeError("Unable to load a TypeMap - no namespace file found")


DynamicTable = __TYPE_MAP.get_container_cls(CORE_NAMESPACE, 'DynamicTable')
VectorData = __TYPE_MAP.get_container_cls(CORE_NAMESPACE, 'VectorData')
VectorIndex = __TYPE_MAP.get_container_cls(CORE_NAMESPACE, 'VectorIndex')
ElementIdentifiers = __TYPE_MAP.get_container_cls(CORE_NAMESPACE, 'ElementIdentifiers')
DynamicTableRegion = __TYPE_MAP.get_container_cls(CORE_NAMESPACE, 'DynamicTableRegion')
CSRMatrix = __TYPE_MAP.get_container_cls(CORE_NAMESPACE, 'CSRMatrix')


@docval({'name': 'extensions', 'type': (str, TypeMap, list),
         'doc': 'a path to a namespace, a TypeMap, or a list consisting paths to namespaces and TypeMaps',
         'default': None},
        returns="the namespaces loaded from the given file", rtype=tuple,
        is_method=False)
def get_type_map(**kwargs):
    '''
    Get a BuildManager to use for I/O using the given extensions. If no extensions are provided,
    return a BuildManager that uses the core namespace
    '''
    extensions = getargs('extensions', kwargs)
    type_map = None
    if extensions is None:
        type_map = deepcopy(__TYPE_MAP)
    else:
        if isinstance(extensions, TypeMap):
            type_map = extensions
        else:
            type_map = deepcopy(__TYPE_MAP)
        if isinstance(extensions, list):
            for ext in extensions:
                if isinstance(ext, str):
                    type_map.load_namespaces(ext)
                elif isinstance(ext, TypeMap):
                    type_map.merge(ext)
                else:
                    msg = 'extensions must be a list of paths to namespace specs or a TypeMaps'
                    raise ValueError(msg)
        elif isinstance(extensions, str):
            type_map.load_namespaces(extensions)
        elif isinstance(extensions, TypeMap):
            type_map.merge(extensions)
    return type_map


@docval({'name': 'extensions', 'type': (str, TypeMap, list),
         'doc': 'a path to a namespace, a TypeMap, or a list consisting paths to namespaces and TypeMaps',
         'default': None},
        returns="the namespaces loaded from the given file", rtype=tuple,
        is_method=False)
def get_manager(**kwargs):
    '''
    Get a BuildManager to use for I/O using the given extensions. If no extensions are provided,
    return a BuildManager that uses the core namespace
    '''
    type_map = call_docval_func(get_type_map, kwargs)
    return BuildManager(type_map)


# a function to get the container class for a give type
@docval({'name': 'data_type', 'type': str,
         'doc': 'the data_type to get the Container class for'},
        {'name': 'namespace', 'type': str, 'doc': 'the namespace the data_type is defined in'},
        is_method=False)
def get_class(**kwargs):
    """Get the class object of the Container subclass corresponding to a given neurdata_type.
    """
    data_type, namespace = getargs('data_type', 'namespace', kwargs)
    return __TYPE_MAP.get_container_cls(namespace, data_type)


@docval({'name': 'io', 'type': HDMFIO,
         'doc': 'the HDMFIO object to read from'},
        {'name': 'namespace', 'type': str,
         'doc': 'the namespace to validate against', 'default': CORE_NAMESPACE},
        returns="errors in the file", rtype=list,
        is_method=False)
def validate(**kwargs):
    """Validate an file against a namespace"""
    io, namespace = getargs('io', 'namespace', kwargs)
    builder = io.read_builder()
    validator = ValidatorMap(io.manager.namespace_catalog.get_namespace(name=namespace))
    return validator.validate(builder)



class OrphanContainerWarning(UserWarning):
    """
    Raised when a container does not have a parent.

    Only the top level container (e.g. file) should be
    without a parent
    """
    pass


class MissingRequiredWarning(UserWarning):
    """
    Raised when a required field is missing.
    """
    pass



import numpy as np

from hdmf.utils import docval, fmt_docval_args, get_docval, popargs, AllowPositional
from hdmf.testing import TestCase


class MyTestClass(object):

    @docval({'name': 'arg1', 'type': str, 'doc': 'argument1 is a str'})
    def basic_add(self, **kwargs):
        return kwargs

    @docval({'name': 'arg1', 'type': str, 'doc': 'argument1 is a str'},
            {'name': 'arg2', 'type': int, 'doc': 'argument2 is a int'})
    def basic_add2(self, **kwargs):
        return kwargs

    @docval({'name': 'arg1', 'type': str, 'doc': 'argument1 is a str'},
            {'name': 'arg2', 'type': 'int', 'doc': 'argument2 is a int'},
            {'name': 'arg3', 'type': bool, 'doc': 'argument3 is a bool. it defaults to False', 'default': False})
    def basic_add2_kw(self, **kwargs):
        return kwargs

    @docval({'name': 'arg1', 'type': str, 'doc': 'argument1 is a str', 'default': 'a'},
            {'name': 'arg2', 'type': int, 'doc': 'argument2 is a int', 'default': 1})
    def basic_only_kw(self, **kwargs):
        return kwargs

    @docval({'name': 'arg1', 'type': str, 'doc': 'argument1 is a str'},
            {'name': 'arg2', 'type': 'int', 'doc': 'argument2 is a int'},
            {'name': 'arg3', 'type': bool, 'doc': 'argument3 is a bool. it defaults to False', 'default': False},
            allow_extra=True)
    def basic_add2_kw_allow_extra(self, **kwargs):
        return kwargs


class MyTestSubclass(MyTestClass):

    @docval({'name': 'arg1', 'type': str, 'doc': 'argument1 is a str'},
            {'name': 'arg2', 'type': int, 'doc': 'argument2 is a int'})
    def basic_add(self, **kwargs):
        return kwargs

    @docval({'name': 'arg1', 'type': str, 'doc': 'argument1 is a str'},
            {'name': 'arg2', 'type': int, 'doc': 'argument2 is a int'},
            {'name': 'arg3', 'type': bool, 'doc': 'argument3 is a bool. it defaults to False', 'default': False},
            {'name': 'arg4', 'type': str, 'doc': 'argument4 is a str'},
            {'name': 'arg5', 'type': 'float', 'doc': 'argument5 is a float'},
            {'name': 'arg6', 'type': bool, 'doc': 'argument6 is a bool. it defaults to False', 'default': None})
    def basic_add2_kw(self, **kwargs):
        return kwargs


class MyChainClass(MyTestClass):

    @docval({'name': 'arg1', 'type': (str, 'MyChainClass'), 'doc': 'arg1 is a string or MyChainClass'},
            {'name': 'arg2', 'type': ('array_data', 'MyChainClass'),
             'doc': 'arg2 is array data or MyChainClass. it defaults to None', 'default': None},
            {'name': 'arg3', 'type': ('array_data', 'MyChainClass'), 'doc': 'arg3 is array data or MyChainClass',
             'shape': (None, 2)},
            {'name': 'arg4', 'type': ('array_data', 'MyChainClass'),
             'doc': 'arg3 is array data or MyChainClass. it defaults to None.', 'shape': (None, 2), 'default': None})
    def __init__(self, **kwargs):
        self._arg1, self._arg2, self._arg3, self._arg4 = popargs('arg1', 'arg2', 'arg3', 'arg4', kwargs)

    @property
    def arg1(self):
        if isinstance(self._arg1, MyChainClass):
            return self._arg1.arg1
        else:
            return self._arg1

    @property
    def arg2(self):
        if isinstance(self._arg2, MyChainClass):
            return self._arg2.arg2
        else:
            return self._arg2

    @property
    def arg3(self):
        if isinstance(self._arg3, MyChainClass):
            return self._arg3.arg3
        else:
            return self._arg3

    @arg3.setter
    def arg3(self, val):
        self._arg3 = val

    @property
    def arg4(self):
        if isinstance(self._arg4, MyChainClass):
            return self._arg4.arg4
        else:
            return self._arg4

    @arg4.setter
    def arg4(self, val):
        self._arg4 = val


class TestDocValidator(TestCase):

    def setUp(self):
        self.test_obj = MyTestClass()
        self.test_obj_sub = MyTestSubclass()

    def test_bad_type(self):
        exp_msg = (r"error parsing 'arg1' argument' : argtype must be a type, "
                   r"a str, a list, a tuple, or None - got <class|type 'dict'>")
        with self.assertRaisesRegex(Exception, exp_msg):
            @docval({'name': 'arg1', 'type': {'a': 1}, 'doc': 'this is a bad type'})
            def method(self, **kwargs):
                pass
            method(self, arg1=1234560)

    def test_bad_shape(self):
        @docval({'name': 'arg1', 'type': 'array_data', 'doc': 'this is a bad shape', 'shape': (None, 2)})
        def method(self, **kwargs):
            pass
        with self.assertRaises(ValueError):
            method(self, arg1=[[1]])
        with self.assertRaises(ValueError):
            method(self, arg1=[1])
        # this should work
        method(self, arg1=[[1, 1]])

    def test_multi_shape(self):
        @docval({'name': 'arg1', 'type': 'array_data', 'doc': 'this is a bad shape',
                 'shape': ((None,), (None, 2))})
        def method1(self, **kwargs):
            pass

        method1(self, arg1=[[1, 1]])
        method1(self, arg1=[1, 2])
        with self.assertRaises(ValueError):
            method1(self, arg1=[[1, 1, 1]])

    def test_fmt_docval_args(self):
        """ Test that fmt_docval_args works """
        test_kwargs = {
            'arg1': 'a string',
            'arg2': 1,
            'arg3': True,
        }
        rec_args, rec_kwargs = fmt_docval_args(self.test_obj.basic_add2_kw, test_kwargs)
        exp_args = ['a string', 1]
        self.assertListEqual(rec_args, exp_args)
        exp_kwargs = {'arg3': True}
        self.assertDictEqual(rec_kwargs, exp_kwargs)

    def test_fmt_docval_args_no_docval(self):
        """ Test that fmt_docval_args raises an error when run on function without docval """
        def method1(self, **kwargs):
            pass

        with self.assertRaisesRegex(ValueError, r"no docval found on .*method1.*"):
            fmt_docval_args(method1, {})

    def test_fmt_docval_args_allow_extra(self):
        """ Test that fmt_docval_args works """
        test_kwargs = {
            'arg1': 'a string',
            'arg2': 1,
            'arg3': True,
            'hello': 'abc',
            'list': ['abc', 1, 2, 3]
        }
        rec_args, rec_kwargs = fmt_docval_args(self.test_obj.basic_add2_kw_allow_extra, test_kwargs)
        exp_args = ['a string', 1]
        self.assertListEqual(rec_args, exp_args)
        exp_kwargs = {'arg3': True, 'hello': 'abc', 'list': ['abc', 1, 2, 3]}
        self.assertDictEqual(rec_kwargs, exp_kwargs)

    def test_docval_add(self):
        """Test that docval works with a single positional
           argument
        """
        kwargs = self.test_obj.basic_add('a string')
        self.assertDictEqual(kwargs, {'arg1': 'a string'})

    def test_docval_add_kw(self):
        """Test that docval works with a single positional
           argument passed as key-value
        """
        kwargs = self.test_obj.basic_add(arg1='a string')
        self.assertDictEqual(kwargs, {'arg1': 'a string'})

    def test_docval_add_missing_args(self):
        """Test that docval catches missing argument
           with a single positional argument
        """
        with self.assertRaisesWith(TypeError, "MyTestClass.basic_add: missing argument 'arg1'"):
            self.test_obj.basic_add()

    def test_docval_add2(self):
        """Test that docval works with two positional
           arguments
        """
        kwargs = self.test_obj.basic_add2('a string', 100)
        self.assertDictEqual(kwargs, {'arg1': 'a string', 'arg2': 100})

    def test_docval_add2_w_unicode(self):
        """Test that docval works with two positional
           arguments
        """
        kwargs = self.test_obj.basic_add2(u'a string', 100)
        self.assertDictEqual(kwargs, {'arg1': u'a string', 'arg2': 100})

    def test_docval_add2_kw_default(self):
        """Test that docval works with two positional
           arguments and a keyword argument when using
           default keyword argument value
        """
        kwargs = self.test_obj.basic_add2_kw('a string', 100)
        self.assertDictEqual(kwargs, {'arg1': 'a string', 'arg2': 100, 'arg3': False})

    def test_docval_add2_pos_as_kw(self):
        """Test that docval works with two positional
           arguments and a keyword argument when using
           default keyword argument value, but pass
           positional arguments by key-value
        """
        kwargs = self.test_obj.basic_add2_kw(arg1='a string', arg2=100)
        self.assertDictEqual(kwargs, {'arg1': 'a string', 'arg2': 100, 'arg3': False})

    def test_docval_add2_kw_kw_syntax(self):
        """Test that docval works with two positional
           arguments and a keyword argument when specifying
           keyword argument value with keyword syntax
        """
        kwargs = self.test_obj.basic_add2_kw('a string', 100, arg3=True)
        self.assertDictEqual(kwargs, {'arg1': 'a string', 'arg2': 100, 'arg3': True})

    def test_docval_add2_kw_all_kw_syntax(self):
        """Test that docval works with two positional
           arguments and a keyword argument when specifying
           all arguments by key-value
        """
        kwargs = self.test_obj.basic_add2_kw(arg1='a string', arg2=100, arg3=True)
        self.assertDictEqual(kwargs, {'arg1': 'a string', 'arg2': 100, 'arg3': True})

    def test_docval_add2_kw_pos_syntax(self):
        """Test that docval works with two positional
           arguments and a keyword argument when specifying
           keyword argument value with positional syntax
        """
        kwargs = self.test_obj.basic_add2_kw('a string', 100, True)
        self.assertDictEqual(kwargs, {'arg1': 'a string', 'arg2': 100, 'arg3': True})

    def test_docval_add2_kw_pos_syntax_missing_args(self):
        """Test that docval catches incorrect type with two positional
           arguments and a keyword argument when specifying
           keyword argument value with positional syntax
        """
        msg = "MyTestClass.basic_add2_kw: incorrect type for 'arg2' (got 'str', expected 'int')"
        with self.assertRaisesWith(TypeError, msg):
            self.test_obj.basic_add2_kw('a string', 'bad string')

    def test_docval_add_sub(self):
        """Test that docval works with a two positional arguments,
           where the second is specified by the subclass implementation
        """
        kwargs = self.test_obj_sub.basic_add('a string', 100)
        expected = {'arg1': 'a string', 'arg2': 100}
        self.assertDictEqual(kwargs, expected)

    def test_docval_add2_kw_default_sub(self):
        """Test that docval works with a four positional arguments and
           two keyword arguments, where two positional and one keyword
           argument is specified in both the parent and sublcass implementations
        """
        kwargs = self.test_obj_sub.basic_add2_kw('a string', 100, 'another string', 200.0)
        expected = {'arg1': 'a string', 'arg2': 100,
                    'arg4': 'another string', 'arg5': 200.0,
                    'arg3': False, 'arg6': None}
        self.assertDictEqual(kwargs, expected)

    def test_docval_add2_kw_default_sub_missing_args(self):
        """Test that docval catches missing arguments with a four positional arguments
           and two keyword arguments, where two positional and one keyword
           argument is specified in both the parent and sublcass implementations,
           when using default values for keyword arguments
        """
        with self.assertRaisesWith(TypeError, "MyTestSubclass.basic_add2_kw: missing argument 'arg5'"):
            self.test_obj_sub.basic_add2_kw('a string', 100, 'another string')

    def test_docval_add2_kw_kwsyntax_sub(self):
        """Test that docval works when called with a four positional
           arguments and two keyword arguments, where two positional
           and one keyword argument is specified in both the parent
           and sublcass implementations
        """
        kwargs = self.test_obj_sub.basic_add2_kw('a string', 100, 'another string', 200.0, arg6=True)
        expected = {'arg1': 'a string', 'arg2': 100,
                    'arg4': 'another string', 'arg5': 200.0,
                    'arg3': False, 'arg6': True}
        self.assertDictEqual(kwargs, expected)

    def test_docval_add2_kw_kwsyntax_sub_missing_args(self):
        """Test that docval catches missing arguments when called with a four positional
           arguments and two keyword arguments, where two positional and one keyword
           argument is specified in both the parent and sublcass implementations
        """
        with self.assertRaisesWith(TypeError, "MyTestSubclass.basic_add2_kw: missing argument 'arg5'"):
            self.test_obj_sub.basic_add2_kw('a string', 100, 'another string', arg6=True)

    def test_docval_add2_kw_kwsyntax_sub_nonetype_arg(self):
        """Test that docval catches NoneType when called with a four positional
           arguments and two keyword arguments, where two positional and one keyword
           argument is specified in both the parent and sublcass implementations
        """
        msg = "MyTestSubclass.basic_add2_kw: None is not allowed for 'arg5' (expected 'float', not None)"
        with self.assertRaisesWith(TypeError, msg):
            self.test_obj_sub.basic_add2_kw('a string', 100, 'another string', None, arg6=True)

    def test_only_kw_no_args(self):
        """Test that docval parses arguments when only keyword
           arguments exist, and no arguments are specified
        """
        kwargs = self.test_obj.basic_only_kw()
        self.assertDictEqual(kwargs, {'arg1': 'a', 'arg2': 1})

    def test_only_kw_arg1_no_arg2(self):
        """Test that docval parses arguments when only keyword
           arguments exist, and only first argument is specified
           as key-value
        """
        kwargs = self.test_obj.basic_only_kw(arg1='b')
        self.assertDictEqual(kwargs, {'arg1': 'b', 'arg2': 1})

    def test_only_kw_arg1_pos_no_arg2(self):
        """Test that docval parses arguments when only keyword
           arguments exist, and only first argument is specified
           as positional argument
        """
        kwargs = self.test_obj.basic_only_kw('b')
        self.assertDictEqual(kwargs, {'arg1': 'b', 'arg2': 1})

    def test_only_kw_arg2_no_arg1(self):
        """Test that docval parses arguments when only keyword
           arguments exist, and only second argument is specified
           as key-value
        """
        kwargs = self.test_obj.basic_only_kw(arg2=2)
        self.assertDictEqual(kwargs, {'arg1': 'a', 'arg2': 2})

    def test_only_kw_arg1_arg2(self):
        """Test that docval parses arguments when only keyword
           arguments exist, and both arguments are specified
           as key-value
        """
        kwargs = self.test_obj.basic_only_kw(arg1='b', arg2=2)
        self.assertDictEqual(kwargs, {'arg1': 'b', 'arg2': 2})

    def test_only_kw_arg1_arg2_pos(self):
        """Test that docval parses arguments when only keyword
           arguments exist, and both arguments are specified
           as positional arguments
        """
        kwargs = self.test_obj.basic_only_kw('b', 2)
        self.assertDictEqual(kwargs, {'arg1': 'b', 'arg2': 2})

    def test_extra_kwarg(self):
        """Test that docval parses arguments when only keyword
           arguments exist, and both arguments are specified
           as positional arguments
        """
        with self.assertRaises(TypeError):
            self.test_obj.basic_add2_kw('a string', 100, bar=1000)

    def test_extra_args_pos_only(self):
        """Test that docval raises an error if too many positional
           arguments are specified
        """
        msg = ("MyTestClass.basic_add2_kw: Expected at most 3 arguments ['arg1', 'arg2', 'arg3'], got 4: 4 positional "
               "and 0 keyword []")
        with self.assertRaisesWith(TypeError, msg):
            self.test_obj.basic_add2_kw('a string', 100, True, 'extra')

    def test_extra_args_pos_kw(self):
        """Test that docval raises an error if too many positional
           arguments are specified and a keyword arg is specified
        """
        msg = ("MyTestClass.basic_add2_kw: Expected at most 3 arguments ['arg1', 'arg2', 'arg3'], got 4: 3 positional "
               "and 1 keyword ['arg3']")
        with self.assertRaisesWith(TypeError, msg):
            self.test_obj.basic_add2_kw('a string', 'extra', 100, arg3=True)

    def test_extra_kwargs_pos_kw(self):
        """Test that docval raises an error if extra keyword
           arguments are specified
        """
        msg = ("MyTestClass.basic_add2_kw: Expected at most 3 arguments ['arg1', 'arg2', 'arg3'], got 4: 2 positional "
               "and 2 keyword ['arg3', 'extra']")
        with self.assertRaisesWith(TypeError, msg):
            self.test_obj.basic_add2_kw('a string', 100, extra='extra', arg3=True)

    def test_extra_args_pos_only_ok(self):
        """Test that docval raises an error if too many positional
           arguments are specified even if allow_extra is True
        """
        msg = ("MyTestClass.basic_add2_kw_allow_extra: Expected at most 3 arguments ['arg1', 'arg2', 'arg3'], got "
               "4 positional")
        with self.assertRaisesWith(TypeError, msg):
            self.test_obj.basic_add2_kw_allow_extra('a string', 100, True, 'extra', extra='extra')

    def test_extra_args_pos_kw_ok(self):
        """Test that docval does not raise an error if too many
           keyword arguments are specified and allow_extra is True
        """
        kwargs = self.test_obj.basic_add2_kw_allow_extra('a string', 100, True, extra='extra')
        self.assertDictEqual(kwargs, {'arg1': 'a string', 'arg2': 100, 'arg3': True, 'extra': 'extra'})

    def test_dup_kw(self):
        """Test that docval raises an error if a keyword argument
           captures a positional argument before all positional
           arguments have been resolved
        """
        with self.assertRaisesWith(TypeError, "MyTestClass.basic_add2_kw: got multiple values for argument 'arg1'"):
            self.test_obj.basic_add2_kw('a string', 100, arg1='extra')

    def test_extra_args_dup_kw(self):
        """Test that docval raises an error if a keyword argument
           captures a positional argument before all positional
           arguments have been resolved and allow_extra is True
        """
        msg = "MyTestClass.basic_add2_kw_allow_extra: got multiple values for argument 'arg1'"
        with self.assertRaisesWith(TypeError, msg):
            self.test_obj.basic_add2_kw_allow_extra('a string', 100, True, arg1='extra')

    def test_unsupported_docval_term(self):
        """Test that docval does not allow setting of arguments
           marked as unsupported
        """
        @docval({'name': 'arg1', 'type': 'array_data', 'doc': 'this is a bad shape', 'unsupported': 'hi!'})
        def method(self, **kwargs):
            pass
        with self.assertRaises(ValueError):
            method(self, arg1=[[1, 1]])

    def test_catch_dup_names(self):
        """Test that docval does not allow duplicate argument names
        """
        @docval({'name': 'arg1', 'type': 'array_data', 'doc': 'this is a bad shape'},
                {'name': 'arg1', 'type': 'array_data', 'doc': 'this is a bad shape2'})
        def method(self, **kwargs):
            pass
        msg = "TestDocValidator.test_catch_dup_names.<locals>.method: The following names are duplicated: ['arg1']"
        with self.assertRaisesWith(ValueError, msg):
            method(self, arg1=[1])

    def test_get_docval_all(self):
        """Test that get_docval returns a tuple of the docval arguments
        """
        args = get_docval(self.test_obj.basic_add2)
        self.assertTupleEqual(args, ({'name': 'arg1', 'type': str, 'doc': 'argument1 is a str'},
                                     {'name': 'arg2', 'type': int, 'doc': 'argument2 is a int'}))

    def test_get_docval_one_arg(self):
        """Test that get_docval returns the matching docval argument
        """
        arg = get_docval(self.test_obj.basic_add2, 'arg2')
        self.assertTupleEqual(arg, ({'name': 'arg2', 'type': int, 'doc': 'argument2 is a int'},))

    def test_get_docval_two_args(self):
        """Test that get_docval returns the matching docval arguments in order
        """
        args = get_docval(self.test_obj.basic_add2, 'arg2', 'arg1')
        self.assertTupleEqual(args, ({'name': 'arg2', 'type': int, 'doc': 'argument2 is a int'},
                                     {'name': 'arg1', 'type': str, 'doc': 'argument1 is a str'}))

    def test_get_docval_missing_arg(self):
        """Test that get_docval throws error if the matching docval argument is not found
        """
        with self.assertRaisesWith(ValueError, "Function basic_add2 does not have docval argument 'arg3'"):
            get_docval(self.test_obj.basic_add2, 'arg3')

    def test_get_docval_missing_args(self):
        """Test that get_docval throws error if the matching docval arguments is not found
        """
        with self.assertRaisesWith(ValueError, "Function basic_add2 does not have docval argument 'arg3'"):
            get_docval(self.test_obj.basic_add2, 'arg3', 'arg4')

    def test_get_docval_missing_arg_of_many_ok(self):
        """Test that get_docval throws error if the matching docval arguments is not found
        """
        with self.assertRaisesWith(ValueError, "Function basic_add2 does not have docval argument 'arg3'"):
            get_docval(self.test_obj.basic_add2, 'arg2', 'arg3')

    def test_get_docval_none(self):
        """Test that get_docval returns an empty tuple if there is no docval
        """
        args = get_docval(self.test_obj.__init__)
        self.assertTupleEqual(args, tuple())

    def test_get_docval_none_arg(self):
        """Test that get_docval throws error if there is no docval and an argument name is passed
        """
        with self.assertRaisesWith(ValueError, 'Function __init__ has no docval arguments'):
            get_docval(self.test_obj.__init__, 'arg3')

    def test_bool_type(self):
        @docval({'name': 'arg1', 'type': bool, 'doc': 'this is a bool'})
        def method(self, **kwargs):
            return popargs('arg1', kwargs)

        res = method(self, arg1=True)
        self.assertEqual(res, True)
        self.assertIsInstance(res, bool)

        res = method(self, arg1=np.bool_(True))
        self.assertEqual(res, np.bool_(True))
        self.assertIsInstance(res, np.bool_)

    def test_bool_string_type(self):
        @docval({'name': 'arg1', 'type': 'bool', 'doc': 'this is a bool'})
        def method(self, **kwargs):
            return popargs('arg1', kwargs)

        res = method(self, arg1=True)
        self.assertEqual(res, True)
        self.assertIsInstance(res, bool)

        res = method(self, arg1=np.bool_(True))
        self.assertEqual(res, np.bool_(True))
        self.assertIsInstance(res, np.bool_)

    def test_allow_positional_warn(self):
        @docval({'name': 'arg1', 'type': bool, 'doc': 'this is a bool'}, allow_positional=AllowPositional.WARNING)
        def method(self, **kwargs):
            return popargs('arg1', kwargs)

        # check that supplying a keyword arg is OK
        res = method(self, arg1=True)
        self.assertEqual(res, True)
        self.assertIsInstance(res, bool)

        # check that supplying a positional arg raises a warning
        msg = ('TestDocValidator.test_allow_positional_warn.<locals>.method: '
               'Positional arguments are discouraged and may be forbidden in a future release.')
        with self.assertWarnsWith(FutureWarning, msg):
            method(self, True)

    def test_allow_positional_error(self):
        @docval({'name': 'arg1', 'type': bool, 'doc': 'this is a bool'}, allow_positional=AllowPositional.ERROR)
        def method(self, **kwargs):
            return popargs('arg1', kwargs)

        # check that supplying a keyword arg is OK
        res = method(self, arg1=True)
        self.assertEqual(res, True)
        self.assertIsInstance(res, bool)

        # check that supplying a positional arg raises an error
        msg = ('TestDocValidator.test_allow_positional_error.<locals>.method: '
               'Only keyword arguments (e.g., func(argname=value, ...)) are allowed.')
        with self.assertRaisesWith(SyntaxError, msg):
            method(self, True)


class TestDocValidatorChain(TestCase):

    def setUp(self):
        self.obj1 = MyChainClass('base', [[1, 2], [3, 4], [5, 6]], [[10, 20]])
        # note that self.obj1.arg3 == [[1, 2], [3, 4], [5, 6]]

    def test_type_arg(self):
        """Test that passing an object for an argument that allows a specific type works"""
        obj2 = MyChainClass(self.obj1, [[10, 20], [30, 40], [50, 60]], [[10, 20]])
        self.assertEqual(obj2.arg1, 'base')

    def test_type_arg_wrong_type(self):
        """Test that passing an object for an argument that does not match a specific type raises an error"""
        err_msg = "MyChainClass.__init__: incorrect type for 'arg1' (got 'object', expected 'str or MyChainClass')"
        with self.assertRaisesWith(TypeError, err_msg):
            MyChainClass(object(), [[10, 20], [30, 40], [50, 60]], [[10, 20]])

    def test_shape_valid_unpack(self):
        """Test that passing an object for an argument with required shape tests the shape of object.argument"""
        obj2 = MyChainClass(self.obj1, [[10, 20], [30, 40], [50, 60]], [[10, 20]])
        obj3 = MyChainClass(self.obj1, obj2, [[100, 200]])
        self.assertListEqual(obj3.arg3, obj2.arg3)

    def test_shape_invalid_unpack(self):
        """Test that passing an object for an argument with required shape and object.argument has an invalid shape
        raises an error"""
        obj2 = MyChainClass(self.obj1, [[10, 20], [30, 40], [50, 60]], [[10, 20]])
        # change arg3 of obj2 to fail the required shape - contrived, but could happen because datasets can change
        # shape after an object is initialized
        obj2.arg3 = [10, 20, 30]

        err_msg = "MyChainClass.__init__: incorrect shape for 'arg3' (got '(3,)', expected '(None, 2)')"
        with self.assertRaisesWith(ValueError, err_msg):
            MyChainClass(self.obj1, obj2, [[100, 200]])

    def test_shape_none_unpack(self):
        """Test that passing an object for an argument with required shape and object.argument is None is OK"""
        obj2 = MyChainClass(self.obj1, [[10, 20], [30, 40], [50, 60]], [[10, 20]])
        obj2.arg3 = None
        obj3 = MyChainClass(self.obj1, obj2, [[100, 200]])
        self.assertIsNone(obj3.arg3)

    def test_shape_other_unpack(self):
        """Test that passing an object for an argument with required shape and object.argument is an object without
        an argument attribute raises an error"""
        obj2 = MyChainClass(self.obj1, [[10, 20], [30, 40], [50, 60]], [[10, 20]])
        obj2.arg3 = object()

        err_msg = (r"cannot check shape of object '<object object at .*>' for argument 'arg3' "
                   r"\(expected shape '\(None, 2\)'\)")
        with self.assertRaisesRegex(ValueError, err_msg):
            MyChainClass(self.obj1, obj2, [[100, 200]])

    def test_shape_valid_unpack_default(self):
        """Test that passing an object for an argument with required shape and a default value tests the shape of
        object.argument"""
        obj2 = MyChainClass(self.obj1, [[10, 20], [30, 40], [50, 60]], arg4=[[10, 20]])
        obj3 = MyChainClass(self.obj1, [[100, 200], [300, 400], [500, 600]], arg4=obj2)
        self.assertListEqual(obj3.arg4, obj2.arg4)

    def test_shape_invalid_unpack_default(self):
        """Test that passing an object for an argument with required shape and a default value and object.argument has
        an invalid shape raises an error"""
        obj2 = MyChainClass(self.obj1, [[10, 20], [30, 40], [50, 60]], arg4=[[10, 20]])
        # change arg3 of obj2 to fail the required shape - contrived, but could happen because datasets can change
        # shape after an object is initialized
        obj2.arg4 = [10, 20, 30]

        err_msg = "MyChainClass.__init__: incorrect shape for 'arg4' (got '(3,)', expected '(None, 2)')"
        with self.assertRaisesWith(ValueError, err_msg):
            MyChainClass(self.obj1, [[100, 200], [300, 400], [500, 600]], arg4=obj2)

    def test_shape_none_unpack_default(self):
        """Test that passing an object for an argument with required shape and a default value and object.argument is
        an object without an argument attribute raises an error"""
        obj2 = MyChainClass(self.obj1, [[10, 20], [30, 40], [50, 60]], arg4=[[10, 20]])
        # change arg3 of obj2 to fail the required shape - contrived, but could happen because datasets can change
        # shape after an object is initialized
        obj2.arg4 = None
        obj3 = MyChainClass(self.obj1, [[100, 200], [300, 400], [500, 600]], arg4=obj2)
        self.assertIsNone(obj3.arg4)

    def test_shape_other_unpack_default(self):
        """Test that passing an object for an argument with required shape and a default value and object.argument is
        None is OK"""
        obj2 = MyChainClass(self.obj1, [[10, 20], [30, 40], [50, 60]], arg4=[[10, 20]])
        # change arg3 of obj2 to fail the required shape - contrived, but could happen because datasets can change
        # shape after an object is initialized
        obj2.arg4 = object()

        err_msg = (r"cannot check shape of object '<object object at .*>' for argument 'arg4' "
                   r"\(expected shape '\(None, 2\)'\)")
        with self.assertRaisesRegex(ValueError, err_msg):
            MyChainClass(self.obj1, [[100, 200], [300, 400], [500, 600]], arg4=obj2)



from hdmf.spec import AttributeSpec, DatasetSpec, SpecCatalog, SpecNamespace, NamespaceCatalog
from hdmf.build import DatasetBuilder, ObjectMapper, BuildManager, TypeMap
from hdmf import Data
from hdmf.utils import docval, getargs
from hdmf.testing import TestCase

import h5py
import numpy as np
import os

from tests.unit.utils import CORE_NAMESPACE


class Baz(Data):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this Baz'},
            {'name': 'data', 'type': (list, h5py.Dataset), 'doc': 'some data'},
            {'name': 'baz_attr', 'type': str, 'doc': 'an attribute'})
    def __init__(self, **kwargs):
        name, data, baz_attr = getargs('name', 'data', 'baz_attr', kwargs)
        super().__init__(name=name, data=data)
        self.__baz_attr = baz_attr

    @property
    def baz_attr(self):
        return self.__baz_attr


class TestDataMap(TestCase):

    def setUp(self):
        self.setUpBazSpec()
        self.spec_catalog = SpecCatalog()
        self.spec_catalog.register_spec(self.baz_spec, 'test.yaml')
        self.namespace = SpecNamespace('a test namespace', CORE_NAMESPACE, [{'source': 'test.yaml'}],
                                       version='0.1.0',
                                       catalog=self.spec_catalog)
        self.namespace_catalog = NamespaceCatalog()
        self.namespace_catalog.add_namespace(CORE_NAMESPACE, self.namespace)
        self.type_map = TypeMap(self.namespace_catalog)
        self.type_map.register_container_type(CORE_NAMESPACE, 'Baz', Baz)
        self.type_map.register_map(Baz, ObjectMapper)
        self.manager = BuildManager(self.type_map)
        self.mapper = ObjectMapper(self.baz_spec)

    def setUpBazSpec(self):
        self.baz_spec = DatasetSpec('an Baz type', 'int', name='MyBaz', data_type_def='Baz',
                                    attributes=[AttributeSpec('baz_attr', 'an example string attribute', 'text')])

    def test_build(self):
        ''' Test default mapping functionality when no attributes are nested '''
        container = Baz('my_baz', list(range(10)), 'abcdefghijklmnopqrstuvwxyz')
        builder = self.mapper.build(container, self.manager)
        expected = DatasetBuilder('my_baz', list(range(10)), attributes={'baz_attr': 'abcdefghijklmnopqrstuvwxyz'})
        self.assertDictEqual(builder, expected)

    def test_append(self):
        with h5py.File('test.h5', 'w') as file:
            test_ds = file.create_dataset('test_ds', data=[1, 2, 3], chunks=True, maxshape=(None,))
            container = Baz('my_baz', test_ds, 'abcdefghijklmnopqrstuvwxyz')
            container.append(4)
            np.testing.assert_array_equal(container[:], [1, 2, 3, 4])
        os.remove('test.h5')

    def test_extend(self):
        with h5py.File('test.h5', 'w') as file:
            test_ds = file.create_dataset('test_ds', data=[1, 2, 3], chunks=True, maxshape=(None,))
            container = Baz('my_baz', test_ds, 'abcdefghijklmnopqrstuvwxyz')
            container.extend([4, 5])
            np.testing.assert_array_equal(container[:], [1, 2, 3, 4, 5])
        os.remove('test.h5')



import copy as _copy
from abc import ABCMeta
import collections
import h5py
import numpy as np
import warnings
from enum import Enum

__macros = {
    'array_data': [np.ndarray, list, tuple, h5py.Dataset],
    'scalar_data': [str, int, float],
}

# code to signify how to handle positional arguments in docval
AllowPositional = Enum('AllowPositional', 'ALLOWED WARNING ERROR')


def docval_macro(macro):
    """Class decorator to add the class to a list of types associated with the key macro in the __macros dict
    """
    def _dec(cls):
        if macro not in __macros:
            __macros[macro] = list()
        __macros[macro].append(cls)
        return cls
    return _dec


def __type_okay(value, argtype, allow_none=False):
    """Check a value against a type

       The difference between this function and :py:func:`isinstance` is that
       it allows specifying a type as a string. Furthermore, strings allow for specifying more general
       types, such as a simple numeric type (i.e. ``argtype``="num").

       Args:
           value (any): the value to check
           argtype (type, str): the type to check for
           allow_none (bool): whether or not to allow None as a valid value


       Returns:
           bool: True if value is a valid instance of argtype
    """
    if value is None:
        return allow_none
    if isinstance(argtype, str):
        if argtype in __macros:
            return __type_okay(value, __macros[argtype], allow_none=allow_none)
        elif argtype == 'int':
            return __is_int(value)
        elif argtype == 'float':
            return __is_float(value)
        elif argtype == 'bool':
            return __is_bool(value)
        return argtype in [cls.__name__ for cls in value.__class__.__mro__]
    elif isinstance(argtype, type):
        if argtype is int:
            return __is_int(value)
        elif argtype is float:
            return __is_float(value)
        elif argtype is bool:
            return __is_bool(value)
        return isinstance(value, argtype)
    elif isinstance(argtype, tuple) or isinstance(argtype, list):
        return any(__type_okay(value, i) for i in argtype)
    else:    # argtype is None
        return True


def __shape_okay_multi(value, argshape):
    if type(argshape[0]) in (tuple, list):  # if multiple shapes are present
        return any(__shape_okay(value, a) for a in argshape)
    else:
        return __shape_okay(value, argshape)


def __shape_okay(value, argshape):
    valshape = get_data_shape(value)
    if not len(valshape) == len(argshape):
        return False
    for a, b in zip(valshape, argshape):
        if b not in (a, None):
            return False
    return True


def __is_int(value):
    return any(isinstance(value, i) for i in (int, np.int8, np.int16, np.int32, np.int64))


def __is_float(value):
    SUPPORTED_FLOAT_TYPES = [float, np.float16, np.float32, np.float64]
    if hasattr(np, "float128"):
        SUPPORTED_FLOAT_TYPES.append(np.float128)
    if hasattr(np, "longdouble"):
        # on windows python<=3.5, h5py floats resolve float64s as either np.float64 or np.longdouble
        # non-deterministically. a future version of h5py will fix this. see #112
        SUPPORTED_FLOAT_TYPES.append(np.longdouble)
    return any(isinstance(value, i) for i in SUPPORTED_FLOAT_TYPES)


def __is_bool(value):
    return isinstance(value, bool) or isinstance(value, np.bool_)


def __format_type(argtype):
    if isinstance(argtype, str):
        return argtype
    elif isinstance(argtype, type):
        return argtype.__name__
    elif isinstance(argtype, tuple) or isinstance(argtype, list):
        types = [__format_type(i) for i in argtype]
        if len(types) > 1:
            return "%s or %s" % (", ".join(types[:-1]), types[-1])
        else:
            return types[0]
    elif argtype is None:
        return "any type"
    else:
        raise ValueError("argtype must be a type, str, list, or tuple")


def __parse_args(validator, args, kwargs, enforce_type=True, enforce_shape=True, allow_extra=False,  # noqa: C901
                 allow_positional=AllowPositional.ALLOWED):
    """
    Internal helper function used by the docval decorator to parse and validate function arguments

    :param validator: List of dicts from docval with the description of the arguments
    :param args: List of the values of positional arguments supplied by the caller
    :param kwargs: Dict keyword arguments supplied by the caller where keys are the argument name and
                   values are the argument value.
    :param enforce_type: Boolean indicating whether the type of arguments should be enforced
    :param enforce_shape: Boolean indicating whether the dimensions of array arguments
                          should be enforced if possible.
    :param allow_extra: Boolean indicating whether extra keyword arguments are allowed (if False and extra keyword
                        arguments are specified, then an error is raised).
    :param allow_positional: integer code indicating whether positional arguments are allowed:
                             AllowPositional.ALLOWED: positional arguments are allowed
                             AllowPositional.WARNING: return warning if positional arguments are supplied
                             AllowPositional.ERROR: return error if positional arguments are supplied

    :return: Dict with:
        * 'args' : Dict all arguments where keys are the names and values are the values of the arguments.
        * 'errors' : List of string with error messages
    """
    ret = dict()
    syntax_errors = list()
    type_errors = list()
    value_errors = list()
    future_warnings = list()
    argsi = 0
    extras = dict()  # has to be initialized to empty here, to avoid spurious errors reported upon early raises

    try:
        # check for duplicates in docval
        names = [x['name'] for x in validator]
        duplicated = [item for item, count in collections.Counter(names).items()
                      if count > 1]
        if duplicated:
            raise ValueError(
                'The following names are duplicated: {}'.format(duplicated))

        if allow_extra:  # extra keyword arguments are allowed so do not consider them when checking number of args
            if len(args) > len(validator):
                raise TypeError(
                    'Expected at most %d arguments %r, got %d positional' % (len(validator), names, len(args))
                )
        else:  # allow for keyword args
            if len(args) + len(kwargs) > len(validator):
                raise TypeError(
                    'Expected at most %d arguments %r, got %d: %d positional and %d keyword %s'
                    % (len(validator), names, len(args) + len(kwargs), len(args), len(kwargs), sorted(kwargs))
                )

        if args:
            if allow_positional == AllowPositional.WARNING:
                msg = 'Positional arguments are discouraged and may be forbidden in a future release.'
                future_warnings.append(msg)
            elif allow_positional == AllowPositional.ERROR:
                msg = 'Only keyword arguments (e.g., func(argname=value, ...)) are allowed.'
                syntax_errors.append(msg)

        # iterate through the docval specification and find a matching value in args / kwargs
        it = iter(validator)
        arg = next(it)

        # catch unsupported keys
        allowable_terms = ('name', 'doc', 'type', 'shape', 'default', 'help')
        unsupported_terms = set(arg.keys()) - set(allowable_terms)
        if unsupported_terms:
            raise ValueError('docval for {}: {} are not supported by docval'.format(arg['name'],
                                                                                    sorted(unsupported_terms)))
        # process positional arguments of the docval specification (no default value)
        extras = dict(kwargs)
        while True:
            if 'default' in arg:
                break
            argname = arg['name']
            argval_set = False
            if argname in kwargs:
                # if this positional arg is specified by a keyword arg and there are remaining positional args that
                # have not yet been matched, then it is undetermined what those positional args match to. thus, raise
                # an error
                if argsi < len(args):
                    type_errors.append("got multiple values for argument '%s'" % argname)
                argval = kwargs.get(argname)
                extras.pop(argname, None)
                argval_set = True
            elif argsi < len(args):
                argval = args[argsi]
                argval_set = True

            if not argval_set:
                type_errors.append("missing argument '%s'" % argname)
            else:
                if enforce_type:
                    if not __type_okay(argval, arg['type']):
                        if argval is None:
                            fmt_val = (argname, __format_type(arg['type']))
                            type_errors.append("None is not allowed for '%s' (expected '%s', not None)" % fmt_val)
                        else:
                            fmt_val = (argname, type(argval).__name__, __format_type(arg['type']))
                            type_errors.append("incorrect type for '%s' (got '%s', expected '%s')" % fmt_val)
                if enforce_shape and 'shape' in arg:
                    valshape = get_data_shape(argval)
                    while valshape is None:
                        if argval is None:
                            break
                        if not hasattr(argval, argname):
                            fmt_val = (argval, argname, arg['shape'])
                            value_errors.append("cannot check shape of object '%s' for argument '%s' "
                                                "(expected shape '%s')" % fmt_val)
                            break
                        # unpack, e.g. if TimeSeries is passed for arg 'data', then TimeSeries.data is checked
                        argval = getattr(argval, argname)
                        valshape = get_data_shape(argval)
                    if valshape is not None and not __shape_okay_multi(argval, arg['shape']):
                        fmt_val = (argname, valshape, arg['shape'])
                        value_errors.append("incorrect shape for '%s' (got '%s', expected '%s')" % fmt_val)
                ret[argname] = argval
            argsi += 1
            arg = next(it)

        # process arguments of the docval specification with a default value
        while True:
            argname = arg['name']
            if argname in kwargs:
                ret[argname] = kwargs.get(argname)
                extras.pop(argname, None)
            elif len(args) > argsi:
                ret[argname] = args[argsi]
                argsi += 1
            else:
                ret[argname] = _copy.deepcopy(arg['default'])
            argval = ret[argname]
            if enforce_type:
                if not __type_okay(argval, arg['type'], arg['default'] is None):
                    if argval is None and arg['default'] is None:
                        fmt_val = (argname, __format_type(arg['type']))
                        type_errors.append("None is not allowed for '%s' (expected '%s', not None)" % fmt_val)
                    else:
                        fmt_val = (argname, type(argval).__name__, __format_type(arg['type']))
                        type_errors.append("incorrect type for '%s' (got '%s', expected '%s')" % fmt_val)
            if enforce_shape and 'shape' in arg and argval is not None:
                valshape = get_data_shape(argval)
                while valshape is None:
                    if argval is None:
                        break
                    if not hasattr(argval, argname):
                        fmt_val = (argval, argname, arg['shape'])
                        value_errors.append("cannot check shape of object '%s' for argument '%s' (expected shape '%s')"
                                            % fmt_val)
                        break
                    # unpack, e.g. if TimeSeries is passed for arg 'data', then TimeSeries.data is checked
                    argval = getattr(argval, argname)
                    valshape = get_data_shape(argval)
                if valshape is not None and not __shape_okay_multi(argval, arg['shape']):
                    fmt_val = (argname, valshape, arg['shape'])
                    value_errors.append("incorrect shape for '%s' (got '%s', expected '%s')" % fmt_val)
            arg = next(it)
    except StopIteration:
        pass
    except TypeError as e:
        type_errors.append(str(e))
    except ValueError as e:
        value_errors.append(str(e))

    if not allow_extra:
        for key in extras.keys():
            type_errors.append("unrecognized argument: '%s'" % key)
    else:
        # TODO: Extras get stripped out if function arguments are composed with fmt_docval_args.
        # allow_extra needs to be tracked on a function so that fmt_docval_args doesn't strip them out
        for key in extras.keys():
            ret[key] = extras[key]
    return {'args': ret, 'future_warnings': future_warnings, 'type_errors': type_errors, 'value_errors': value_errors,
            'syntax_errors': syntax_errors}


docval_idx_name = '__dv_idx__'
docval_attr_name = '__docval__'
__docval_args_loc = 'args'


def get_docval(func, *args):
    '''Get a copy of docval arguments for a function.
    If args are supplied, return only docval arguments with value for 'name' key equal to the args
    '''
    func_docval = getattr(func, docval_attr_name, None)
    if func_docval:
        if args:
            docval_idx = getattr(func, docval_idx_name, None)
            try:
                return tuple(docval_idx[name] for name in args)
            except KeyError as ke:
                raise ValueError('Function %s does not have docval argument %s' % (func.__name__, str(ke)))
        return tuple(func_docval[__docval_args_loc])
    else:
        if args:
            raise ValueError('Function %s has no docval arguments' % func.__name__)
        return tuple()

# def docval_wrap(func, is_method=True):
#    if is_method:
#        @docval(*get_docval(func))
#        def method(self, **kwargs):
#
#            return call_docval_args(func, kwargs)
#        return method
#    else:
#        @docval(*get_docval(func))
#        def static_method(**kwargs):
#            return call_docval_args(func, kwargs)
#        return method


def fmt_docval_args(func, kwargs):
    ''' Separate positional and keyword arguments

    Useful for methods that wrap other methods
    '''
    func_docval = getattr(func, docval_attr_name, None)
    ret_args = list()
    ret_kwargs = dict()
    kwargs_copy = _copy.copy(kwargs)
    if func_docval:
        for arg in func_docval[__docval_args_loc]:
            val = kwargs_copy.pop(arg['name'], None)
            if 'default' in arg:
                if val is not None:
                    ret_kwargs[arg['name']] = val
            else:
                ret_args.append(val)
        if func_docval['allow_extra']:
            ret_kwargs.update(kwargs_copy)
    else:
        raise ValueError('no docval found on %s' % str(func))
    return ret_args, ret_kwargs


def call_docval_func(func, kwargs):
    fargs, fkwargs = fmt_docval_args(func, kwargs)
    return func(*fargs, **fkwargs)


def __resolve_type(t):
    if t is None:
        return t
    if isinstance(t, str):
        if t in __macros:
            return tuple(__macros[t])
        else:
            return t
    elif isinstance(t, type):
        return t
    elif isinstance(t, (list, tuple)):
        ret = list()
        for i in t:
            resolved = __resolve_type(i)
            if isinstance(resolved, tuple):
                ret.extend(resolved)
            else:
                ret.append(resolved)
        return tuple(ret)
    else:
        msg = "argtype must be a type, a str, a list, a tuple, or None - got %s" % type(t)
        raise ValueError(msg)


def docval(*validator, **options):
    '''A decorator for documenting and enforcing type for instance method arguments.

    This decorator takes a list of dictionaries that specify the method parameters. These
    dictionaries are used for enforcing type and building a Sphinx docstring.

    The first arguments are dictionaries that specify the positional
    arguments and keyword arguments of the decorated function. These dictionaries
    must contain the following keys: ``'name'``, ``'type'``, and ``'doc'``. This will define a
    positional argument. To define a keyword argument, specify a default value
    using the key ``'default'``. To validate the dimensions of an input array
    add the optional ``'shape'`` parameter.

    The decorated method must take ``self`` and ``**kwargs`` as arguments.

    When using this decorator, the functions :py:func:`getargs` and
    :py:func:`popargs` can be used for easily extracting arguments from
    kwargs.

    The following code example demonstrates the use of this decorator:

    .. code-block:: python

       @docval({'name': 'arg1':,   'type': str,           'doc': 'this is the first positional argument'},
               {'name': 'arg2':,   'type': int,           'doc': 'this is the second positional argument'},
               {'name': 'kwarg1':, 'type': (list, tuple), 'doc': 'this is a keyword argument', 'default': list()},
               returns='foo object', rtype='Foo'))
       def foo(self, **kwargs):
           arg1, arg2, kwarg1 = getargs('arg1', 'arg2', 'kwarg1', **kwargs)
           ...

    :param enforce_type: Enforce types of input parameters (Default=True)
    :param returns: String describing the return values
    :param rtype: String describing the data type of the return values
    :param is_method: True if this is decorating an instance or class method, False otherwise (Default=True)
    :param enforce_shape: Enforce the dimensions of input arrays (Default=True)
    :param validator: :py:func:`dict` objects specifying the method parameters
    :param options: additional options for documenting and validating method parameters
    '''
    enforce_type = options.pop('enforce_type', True)
    enforce_shape = options.pop('enforce_shape', True)
    returns = options.pop('returns', None)
    rtype = options.pop('rtype', None)
    is_method = options.pop('is_method', True)
    allow_extra = options.pop('allow_extra', False)
    allow_positional = options.pop('allow_positional', True)

    def dec(func):
        _docval = _copy.copy(options)
        _docval['allow_extra'] = allow_extra
        _docval['allow_positional'] = allow_positional
        func.__name__ = _docval.get('func_name', func.__name__)
        func.__doc__ = _docval.get('doc', func.__doc__)
        pos = list()
        kw = list()
        for a in validator:
            try:
                a['type'] = __resolve_type(a['type'])
            except Exception as e:
                msg = "error parsing '%s' argument' : %s" % (a['name'], e.args[0])
                raise Exception(msg)
            if 'default' in a:
                kw.append(a)
            else:
                pos.append(a)
        loc_val = pos+kw
        _docval[__docval_args_loc] = loc_val

        def func_call(*args, **kwargs):
            parsed = __parse_args(
                        loc_val,
                        args[1:] if is_method else args,
                        kwargs,
                        enforce_type=enforce_type,
                        enforce_shape=enforce_shape,
                        allow_extra=allow_extra,
                        allow_positional=allow_positional)

            parse_warnings = parsed.get('future_warnings')
            if parse_warnings:
                msg = '%s: %s' % (func.__qualname__, ', '.join(parse_warnings))
                warnings.warn(msg, FutureWarning)

            for error_type, ExceptionType in (('type_errors', TypeError),
                                              ('value_errors', ValueError),
                                              ('syntax_errors', SyntaxError)):
                parse_err = parsed.get(error_type)
                if parse_err:
                    msg = '%s: %s' % (func.__qualname__, ', '.join(parse_err))
                    raise ExceptionType(msg)

            if is_method:
                return func(args[0], **parsed['args'])
            else:
                return func(**parsed['args'])

        _rtype = rtype
        if isinstance(rtype, type):
            _rtype = rtype.__name__
        docstring = __googledoc(func, _docval[__docval_args_loc], returns=returns, rtype=_rtype)
        docval_idx = {a['name']: a for a in _docval[__docval_args_loc]}  # cache a name-indexed dictionary of args
        setattr(func_call, '__doc__', docstring)
        setattr(func_call, '__name__', func.__name__)
        setattr(func_call, docval_attr_name, _docval)
        setattr(func_call, docval_idx_name, docval_idx)
        setattr(func_call, '__module__', func.__module__)
        return func_call
    return dec


def __sig_arg(argval):
    if 'default' in argval:
        default = argval['default']
        if isinstance(default, str):
            default = "'%s'" % default
        else:
            default = str(default)
        return "%s=%s" % (argval['name'], default)
    else:
        return argval['name']


def __builddoc(func, validator, docstring_fmt, arg_fmt, ret_fmt=None, returns=None, rtype=None):
    '''Generate a Spinxy docstring'''
    def to_str(argtype):
        if isinstance(argtype, type):
            module = argtype.__module__
            name = argtype.__name__

            if module.startswith("h5py") or module.startswith("pandas") or module.startswith("builtins"):
                return ":py:class:`~{name}`".format(name=name)
            else:
                return ":py:class:`~{module}.{name}`".format(name=name, module=module)
        return argtype

    def __sphinx_arg(arg):
        fmt = dict()
        fmt['name'] = arg.get('name')
        fmt['doc'] = arg.get('doc')
        if isinstance(arg['type'], tuple) or isinstance(arg['type'], list):
            fmt['type'] = " or ".join(map(to_str, arg['type']))
        else:
            fmt['type'] = to_str(arg['type'])
        return arg_fmt.format(**fmt)

    sig = "%s(%s)\n\n" % (func.__name__, ", ".join(map(__sig_arg, validator)))
    desc = func.__doc__.strip() if func.__doc__ is not None else ""
    sig += docstring_fmt.format(description=desc, args="\n".join(map(__sphinx_arg, validator)))

    if not (ret_fmt is None or returns is None or rtype is None):
        sig += ret_fmt.format(returns=returns, rtype=rtype)
    return sig


def __sphinxdoc(func, validator, returns=None, rtype=None):
    arg_fmt = (":param {name}: {doc}\n"
               ":type  {name}: {type}")
    docstring_fmt = ("{description}\n\n"
                     "{args}\n")
    ret_fmt = (":returns: {returns}\n"
               ":rtype: {rtype}")
    return __builddoc(func, validator, docstring_fmt, arg_fmt, ret_fmt=ret_fmt, returns=returns, rtype=rtype)


def __googledoc(func, validator, returns=None, rtype=None):
    arg_fmt = "    {name} ({type}): {doc}"
    docstring_fmt = "{description}\n\n"
    if len(validator) > 0:
        docstring_fmt += "Args:\n{args}\n"
    ret_fmt = ("\nReturns:\n"
               "    {rtype}: {returns}")
    return __builddoc(func, validator, docstring_fmt, arg_fmt, ret_fmt=ret_fmt, returns=returns, rtype=rtype)


def getargs(*argnames):
    '''getargs(*argnames, argdict)
    Convenience function to retrieve arguments from a dictionary in batch
    '''
    if len(argnames) < 2:
        raise ValueError('Must supply at least one key and a dict')
    if not isinstance(argnames[-1], dict):
        raise ValueError('last argument must be dict')
    kwargs = argnames[-1]
    if len(argnames) == 2:
        return kwargs.get(argnames[0])
    return [kwargs.get(arg) for arg in argnames[:-1]]


def popargs(*argnames):
    '''popargs(*argnames, argdict)
    Convenience function to retrieve and remove arguments from a dictionary in batch
    '''
    if len(argnames) < 2:
        raise ValueError('Must supply at least one key and a dict')
    if not isinstance(argnames[-1], dict):
        raise ValueError('last argument must be dict')
    kwargs = argnames[-1]
    if len(argnames) == 2:
        return kwargs.pop(argnames[0])
    return [kwargs.pop(arg) for arg in argnames[:-1]]


class ExtenderMeta(ABCMeta):
    """A metaclass that will extend the base class initialization
       routine by executing additional functions defined in
       classes that use this metaclass

       In general, this class should only be used by core developers.
    """

    __preinit = '__preinit'

    @classmethod
    def pre_init(cls, func):
        setattr(func, cls.__preinit, True)
        return classmethod(func)

    __postinit = '__postinit'

    @classmethod
    def post_init(cls, func):
        '''A decorator for defining a routine to run after creation of a type object.

        An example use of this method would be to define a classmethod that gathers
        any defined methods or attributes after the base Python type construction (i.e. after
        :py:func:`type` has been called)
        '''
        setattr(func, cls.__postinit, True)
        return classmethod(func)

    def __init__(cls, name, bases, classdict):
        it = (getattr(cls, n) for n in dir(cls))
        it = (a for a in it if hasattr(a, cls.__preinit))
        for func in it:
            func(name, bases, classdict)
        super().__init__(name, bases, classdict)
        it = (getattr(cls, n) for n in dir(cls))
        it = (a for a in it if hasattr(a, cls.__postinit))
        for func in it:
            func(name, bases, classdict)


def get_data_shape(data, strict_no_data_load=False):
    """
    Helper function used to determine the shape of the given array.

    :param data: Array for which we should determine the shape.
    :type data: List, numpy.ndarray, DataChunkIterator, any object that support __len__ or .shape.
    :param strict_no_data_load: In order to determine the shape of nested tuples and lists, this function
                recursively inspects elements along the dimensions, assuming that the data has a regular,
                rectangular shape. In the case of out-of-core iterators this means that the first item
                along each dimensions would potentially be loaded into memory. By setting this option
                we enforce that this does not happen, at the cost that we may not be able to determine
                the shape of the array.
    :return: Tuple of ints indicating the size of known dimensions. Dimensions for which the size is unknown
             will be set to None.
    """
    def __get_shape_helper(local_data):
        shape = list()
        if hasattr(local_data, '__len__'):
            shape.append(len(local_data))
            if len(local_data) and not isinstance(local_data[0], (str, bytes)):
                shape.extend(__get_shape_helper(local_data[0]))
        return tuple(shape)
    if hasattr(data, 'maxshape'):
        return data.maxshape
    elif hasattr(data, 'shape'):
        return data.shape
    elif isinstance(data, dict):
        return None
    elif hasattr(data, '__len__') and not isinstance(data, (str, bytes)):
        if not strict_no_data_load or (isinstance(data, list) or isinstance(data, tuple) or isinstance(data, set)):
            return __get_shape_helper(data)
        else:
            return None
    else:
        return None


def pystr(s):
    """
    Convert a string of characters to Python str object
    """
    if isinstance(s, bytes):
        return s.decode('utf-8')
    else:
        return s


class LabelledDict(dict):
    """A dict wrapper class with a label and which allows retrieval of values based on an attribute of the values

    For example, if the key attribute is set as 'name' in __init__, then all objects added to the LabelledDict must have
    a 'name' attribute and a particular object in the LabelledDict can be accessed using the syntax ['object_name'] if
    the object.name == 'object_name'. In this way, LabelledDict acts like a set where values can be retrieved using
    square brackets around the value of the key attribute. An 'add' method makes clear the association between the key
    attribute of the LabelledDict and the values of the LabelledDict.

    LabelledDict also supports retrieval of values with the syntax my_dict['attr == val'], which returns a set of
    objects in the LabelledDict which have an attribute 'attr' with a string value 'val'. If no objects match that
    condition, a KeyError is raised. Note that if 'attr' equals the key attribute, then the single matching value is
    returned, not a set.

    Usage:
      LabelledDict(label='my_objects', def_key_name = 'name')
      my_dict[obj.name] = obj
      my_dict.add(obj)  # simpler syntax

    Example:
      # MyTestClass is a class with attributes 'prop1' and 'prop2'. MyTestClass.__init__ sets those attributes.
      ld = LabelledDict(label='all_objects', key_attr='prop1')
      obj1 = MyTestClass('a', 'b')
      obj2 = MyTestClass('d', 'b')
      ld[obj1.prop1] = obj1  # obj1 is added to the LabelledDict with the key obj1.prop1. Any other key is not allowed.
      ld.add(obj2)           # Simpler 'add' syntax enforces the required relationship
      ld['a']                # Returns obj1
      ld['prop1 == a']       # Also returns obj1
      ld['prop2 == b']       # Returns set([obj1, obj2]) - the set of all values v in ld where v.prop2 == 'b'
    """

    @docval({'name': 'label', 'type': str, 'doc': 'the label on this dictionary'},
            {'name': 'key_attr', 'type': str, 'doc': 'the attribute name to use as the key', 'default': 'name'})
    def __init__(self, **kwargs):
        label, key_attr = getargs('label', 'key_attr', kwargs)
        self.__label = label
        self.__key_attr = key_attr

    @property
    def label(self):
        """Return the label of this LabelledDict"""
        return self.__label

    @property
    def key_attr(self):
        """Return the attribute used as the key for values in this LabelledDict"""
        return self.__key_attr

    def __getitem__(self, args):
        """Get a value from the LabelledDict with the given key.

        Supports syntax my_dict['attr == val'], which returns a set of objects in the LabelledDict which have an
        attribute 'attr' with a string value 'val'. If no objects match that condition, a KeyError is raised.

        Note that if 'attr' equals the key attribute, then the single matching value is returned, not a set.
        """
        key = args
        if '==' in args:
            key, val = args.split("==")
            key = key.strip()
            val = val.strip()  # val is a string
            if not key:
                raise ValueError("An attribute name is required before '=='.")
            if not val:
                raise ValueError("A value is required after '=='.")
            if key != self.key_attr:
                ret = set()
                for item in self.values():
                    if getattr(item, key, None) == val:
                        ret.add(item)
                if len(ret):
                    return ret
                else:
                    raise KeyError(val)
            # if key == self.key_attr, then call __getitem__ normally on val
            key = val
        return super().__getitem__(key)

    def __setitem__(self, key, value):
        """Set a value in the LabelledDict with the given key. The key must equal value.key_attr.

        See LabelledDict.add for simpler syntax. Raises ValueError if value does not have attribute key_attr.
        """
        self.__check_value(value)
        if key != getattr(value, self.key_attr):
            raise KeyError("Key '%s' must equal attribute '%s' of '%s'." % (key, self.key_attr, value))
        super().__setitem__(key, value)

    def add(self, value):
        """Add a value to the dict with the key value.key_attr.

        Raises ValueError if value does not have attribute key_attr.
        """
        self.__check_value(value)
        self.__setitem__(getattr(value, self.key_attr), value)

    def __check_value(self, value):
        if not hasattr(value, self.key_attr):
            raise ValueError("Cannot set value '%s' in LabelledDict. Value must have key '%s'."
                             % (value, self.key_attr))



import ruamel.yaml as yaml
import json
import os

from hdmf.spec import AttributeSpec, DatasetSpec, GroupSpec, SpecNamespace, NamespaceCatalog
from hdmf.testing import TestCase


class TestSpecLoad(TestCase):
    NS_NAME = 'test_ns'

    def setUp(self):
        self.attributes = [
            AttributeSpec('attribute1', 'my first attribute', 'text'),
            AttributeSpec('attribute2', 'my second attribute', 'text')
        ]
        self.dset1_attributes = [
            AttributeSpec('attribute3', 'my third attribute', 'text'),
            AttributeSpec('attribute4', 'my fourth attribute', 'text')
        ]
        self.dset2_attributes = [
            AttributeSpec('attribute5', 'my fifth attribute', 'text'),
            AttributeSpec('attribute6', 'my sixth attribute', 'text')
        ]
        self.datasets = [
            DatasetSpec('my first dataset',
                        'int',
                        name='dataset1',
                        attributes=self.dset1_attributes,
                        linkable=True),
            DatasetSpec('my second dataset',
                        'int',
                        name='dataset2',
                        dims=(None, None),
                        attributes=self.dset2_attributes,
                        linkable=True,
                        data_type_def='VoltageArray')
        ]
        self.spec = GroupSpec('A test group',
                              name='root_constructor_datatype',
                              datasets=self.datasets,
                              attributes=self.attributes,
                              linkable=False,
                              data_type_def='EphysData')
        dset1_attributes_ext = [
            AttributeSpec('dset1_extra_attribute', 'an extra attribute for the first dataset', 'text')
        ]
        self.ext_datasets = [
            DatasetSpec('my first dataset extension',
                        'int',
                        name='dataset1',
                        attributes=dset1_attributes_ext,
                        linkable=True),
        ]
        self.ext_attributes = [
            AttributeSpec('ext_extra_attribute', 'an extra attribute for the group', 'text'),
        ]
        self.ext_spec = GroupSpec('A test group extension',
                                  name='root_constructor_datatype',
                                  datasets=self.ext_datasets,
                                  attributes=self.ext_attributes,
                                  linkable=False,
                                  data_type_inc='EphysData',
                                  data_type_def='SpikeData')
        to_dump = {'groups': [self.spec, self.ext_spec]}
        self.specs_path = 'test_load_namespace.specs.yaml'
        self.namespace_path = 'test_load_namespace.namespace.yaml'
        with open(self.specs_path, 'w') as tmp:
            yaml.safe_dump(json.loads(json.dumps(to_dump)), tmp, default_flow_style=False)
        ns_dict = {
            'doc': 'a test namespace',
            'name': self.NS_NAME,
            'schema': [
                {'source': self.specs_path}
            ],
            'version': '0.1.0'
        }
        self.namespace = SpecNamespace.build_namespace(**ns_dict)
        to_dump = {'namespaces': [self.namespace]}
        with open(self.namespace_path, 'w') as tmp:
            yaml.safe_dump(json.loads(json.dumps(to_dump)), tmp, default_flow_style=False)
        self.ns_catalog = NamespaceCatalog()

    def tearDown(self):
        if os.path.exists(self.namespace_path):
            os.remove(self.namespace_path)
        if os.path.exists(self.specs_path):
            os.remove(self.specs_path)

    def test_inherited_attributes(self):
        self.ns_catalog.load_namespaces(self.namespace_path, resolve=True)
        ts_spec = self.ns_catalog.get_spec(self.NS_NAME, 'EphysData')
        es_spec = self.ns_catalog.get_spec(self.NS_NAME, 'SpikeData')
        ts_attrs = {s.name for s in ts_spec.attributes}
        es_attrs = {s.name for s in es_spec.attributes}
        for attr in ts_attrs:
            with self.subTest(attr=attr):
                self.assertIn(attr, es_attrs)
        # self.assertSetEqual(ts_attrs, es_attrs)
        ts_dsets = {s.name for s in ts_spec.datasets}
        es_dsets = {s.name for s in es_spec.datasets}
        for dset in ts_dsets:
            with self.subTest(dset=dset):
                self.assertIn(dset, es_dsets)
        # self.assertSetEqual(ts_dsets, es_dsets)

    def test_inherited_attributes_not_resolved(self):
        self.ns_catalog.load_namespaces(self.namespace_path, resolve=False)
        es_spec = self.ns_catalog.get_spec(self.NS_NAME, 'SpikeData')
        src_attrs = {s.name for s in self.ext_attributes}
        ext_attrs = {s.name for s in es_spec.attributes}
        self.assertSetEqual(src_attrs, ext_attrs)
        src_dsets = {s.name for s in self.ext_datasets}
        ext_dsets = {s.name for s in es_spec.datasets}
        self.assertSetEqual(src_dsets, ext_dsets)


class TestSpecLoadEdgeCase(TestCase):

    def setUp(self):
        self.specs_path = 'test_load_namespace.specs.yaml'
        self.namespace_path = 'test_load_namespace.namespace.yaml'

        # write basically empty specs file
        to_dump = {'groups': []}
        with open(self.specs_path, 'w') as tmp:
            yaml.safe_dump(json.loads(json.dumps(to_dump)), tmp, default_flow_style=False)

    def tearDown(self):
        if os.path.exists(self.namespace_path):
            os.remove(self.namespace_path)
        if os.path.exists(self.specs_path):
            os.remove(self.specs_path)

    def test_build_namespace_missing_version(self):
        """Test that building/creating a SpecNamespace without a version works but raises a warning."""
        # create namespace without version key
        ns_dict = {
            'doc': 'a test namespace',
            'name': 'test_ns',
            'schema': [
                {'source': self.specs_path}
            ],
        }
        msg = ("Loaded namespace 'test_ns' is missing the required key 'version'. Version will be set to "
               "'%s'. Please notify the extension author." % SpecNamespace.UNVERSIONED)
        with self.assertWarnsWith(UserWarning, msg):
            namespace = SpecNamespace.build_namespace(**ns_dict)

        self.assertEqual(namespace.version, SpecNamespace.UNVERSIONED)

    def test_load_namespace_none_version(self):
        """Test that reading a namespace file without a version works but raises a warning."""
        # create namespace with version key (remove it later)
        ns_dict = {
            'doc': 'a test namespace',
            'name': 'test_ns',
            'schema': [
                {'source': self.specs_path}
            ],
            'version': '0.0.1'
        }
        namespace = SpecNamespace.build_namespace(**ns_dict)
        namespace['version'] = None  # work around lack of setter to remove version key

        # write the namespace to file without version key
        to_dump = {'namespaces': [namespace]}
        with open(self.namespace_path, 'w') as tmp:
            yaml.safe_dump(json.loads(json.dumps(to_dump)), tmp, default_flow_style=False)

        # load the namespace from file
        ns_catalog = NamespaceCatalog()
        msg = ("Loaded namespace 'test_ns' is missing the required key 'version'. Version will be set to "
               "'%s'. Please notify the extension author." % SpecNamespace.UNVERSIONED)
        with self.assertWarnsWith(UserWarning, msg):
            ns_catalog.load_namespaces(self.namespace_path)

        self.assertEqual(ns_catalog.get_namespace('test_ns').version, SpecNamespace.UNVERSIONED)

    def test_load_namespace_unversioned_version(self):
        """Test that reading a namespace file with version=unversioned string works but raises a warning."""
        # create namespace with version key (remove it later)
        ns_dict = {
            'doc': 'a test namespace',
            'name': 'test_ns',
            'schema': [
                {'source': self.specs_path}
            ],
            'version': '0.0.1'
        }
        namespace = SpecNamespace.build_namespace(**ns_dict)
        namespace['version'] = str(SpecNamespace.UNVERSIONED)  # work around lack of setter to remove version key

        # write the namespace to file without version key
        to_dump = {'namespaces': [namespace]}
        with open(self.namespace_path, 'w') as tmp:
            yaml.safe_dump(json.loads(json.dumps(to_dump)), tmp, default_flow_style=False)

        # load the namespace from file
        ns_catalog = NamespaceCatalog()
        msg = "Loaded namespace 'test_ns' is unversioned. Please notify the extension author."
        with self.assertWarnsWith(UserWarning, msg):
            ns_catalog.load_namespaces(self.namespace_path)

        self.assertEqual(ns_catalog.get_namespace('test_ns').version, SpecNamespace.UNVERSIONED)

    def test_missing_version_string(self):
        """Test that the constant variable representing a missing version has not changed."""
        self.assertIsNone(SpecNamespace.UNVERSIONED)

    def test_get_namespace_missing_version(self):
        """Test that SpecNamespace.version returns the constant for a missing version if version gets removed."""
        # create namespace with version key (remove it later)
        ns_dict = {
            'doc': 'a test namespace',
            'name': 'test_ns',
            'schema': [
                {'source': self.specs_path}
            ],
            'version': '0.0.1'
        }
        namespace = SpecNamespace.build_namespace(**ns_dict)
        namespace['version'] = None  # work around lack of setter to remove version key

        self.assertEqual(namespace.version, SpecNamespace.UNVERSIONED)
